# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for views."""

import time

import pytest
from flask import url_for
from flask_oauthlib.client import OAuth as FlaskOAuth
from flask_security import login_user
from helpers import check_response_redirect_url
from invenio_accounts.testutils import login_user_via_session
from invenio_db import db
from itsdangerous import TimedJSONWebSignatureSerializer
from mock import MagicMock
from simplejson import JSONDecodeError
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient import InvenioOAuthClient
from invenio_oauthclient._compat import _create_identifier
from invenio_oauthclient.handlers import token_getter
from invenio_oauthclient.models import RemoteAccount, RemoteToken
from invenio_oauthclient.views.client import blueprint as blueprint_client
from invenio_oauthclient.views.client import serializer
from invenio_oauthclient.views.settings import blueprint as blueprint_settings


def mock_response(oauth, remote_app="test", data=None):
    """Mock the oauth response to use the remote."""
    # Mock oauth remote application
    oauth.remote_apps[remote_app].handle_oauth2_response = MagicMock(
        return_value=data
        or {"access_token": "test_access_token", "scope": "", "token_type": "bearer"}
    )


def test_redirect_uri(views_fixture):
    """Test redirect uri."""
    app = views_fixture
    with app.test_client() as client:
        # Test redirect
        resp = client.get(
            url_for(
                "invenio_oauthclient.login",
                remote_app="test",
                next="http://inveniosoftware.org",
            )
        )
        assert resp.status_code == 302

        # Verify parameters
        params = parse_qs(urlparse(resp.location).query)
        assert params["response_type"] == ["code"]
        assert params["client_id"] == ["testid"]
        assert params["redirect_uri"]
        assert params["state"]

        # Verify next parameter in state token does not allow blanco redirects
        state = serializer.loads(params["state"][0])
        assert state["next"] is None

        # Assert redirect uri does not have any parameters.
        params = parse_qs(urlparse(params["redirect_uri"][0]).query)
        assert params == {}

        # Assert that local redirects are allowed
        test_urls = ["/", "/search"]
        for url in test_urls:
            resp = client.get(
                url_for("invenio_oauthclient.login", remote_app="test", next=url)
            )
            check_response_redirect_url(resp, url)

        # Assert that absolute redirects are allowed only if
        # `APP_ALLOWED_HOSTS` is set and includes them. Otherwise, the relative
        # path of the url is extracted and returned. Note if you need to
        # redirect to index page you should pass '/' as next parameter.

        test_url = "http://inveniosoftware.org/test"

        resp = client.get(
            url_for("invenio_oauthclient.login", remote_app="test", next=test_url)
        )

        check_response_redirect_url(resp, urlparse(test_url).path)

        app.config.update({"APP_ALLOWED_HOSTS": ["inveniosoftware.org"]})

        resp = client.get(
            url_for("invenio_oauthclient.login", remote_app="test", next=test_url)
        )

        check_response_redirect_url(resp, test_url)


def test_login(views_fixture):
    """Test login."""
    app = views_fixture
    with app.test_client() as client:
        # Test redirect
        resp = client.get(
            url_for("invenio_oauthclient.login", remote_app="test", next="/")
        )
        assert resp.status_code == 302

        params = parse_qs(urlparse(resp.location).query)
        assert params["response_type"] == ["code"]
        assert params["client_id"] == ["testid"]
        assert params["redirect_uri"]
        assert params["state"]

        # Invalid remote
        resp = client.get(url_for("invenio_oauthclient.login", remote_app="invalid"))
        assert resp.status_code == 404


def test_authorized(base_app, params):
    """Test login."""
    app = base_app

    handled = {}

    def test_authorized_handler(resp, remote, *args, **kwargs):
        """Save configuration."""
        handled["resp"] = resp
        handled["remote"] = remote
        handled["args"] = args
        handled["kwargs"] = kwargs
        return "TEST"

    def test_invalid_authorized_handler(resp, remote, *args, **kwargs):
        """Set wrong configuration."""
        handled["resp"] = 1
        handled["remote"] = 1
        handled["args"] = 1
        handled["kwargs"] = 1

    base_app.config["OAUTHCLIENT_REMOTE_APPS"].update(
        dict(
            test=dict(
                authorized_handler=test_authorized_handler,
                params=params("testid"),
                title="MyLinkedTestAccount",
            ),
            test_invalid=dict(
                authorized_handler=test_invalid_authorized_handler,
                params=params("test_invalidid"),
                title="Test Invalid",
            ),
            full=dict(
                params=params("fullid"),
                title="Full",
            ),
        )
    )

    FlaskOAuth(app)
    InvenioOAuthClient(app)
    base_app.register_blueprint(blueprint_client)
    base_app.register_blueprint(blueprint_settings)

    with app.test_client() as client:
        # Ensure remote apps have been loaded (due to before first
        # request)
        client.get(url_for("invenio_oauthclient.login", remote_app="test"))
        mock_response(app.extensions["oauthlib.client"], "test")
        mock_response(app.extensions["oauthlib.client"], "test_invalid")

        from invenio_oauthclient.views.client import serializer

        state = serializer.dumps(
            {
                "app": "test",
                "sid": _create_identifier(),
                "next": None,
            }
        )

        resp = client.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="test",
                code="test",
                state=state,
            )
        )
        assert resp.data == b"TEST"
        assert handled["remote"].name == "test"
        assert not handled["args"]
        assert not handled["kwargs"]
        assert handled["resp"]["access_token"] == "test_access_token"

        state = serializer.dumps(
            {
                "app": "test_invalid",
                "sid": _create_identifier(),
                "next": None,
            }
        )

        # handler should return something
        # Flask>1.0 is throwing TypeError and Flask<1.0 ValueError
        with pytest.raises((ValueError, TypeError)):
            client.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="test_invalid",
                    code="test",
                    state=state,
                )
            )


def test_invalid_authorized_response(views_fixture):
    """Test login."""
    app = views_fixture
    oauth = app.extensions["oauthlib.client"]
    with app.test_client() as client:
        # Fake an authorized request
        # Ensure remote apps have been loaded (due to before first
        # request)
        client.get(url_for("invenio_oauthclient.login", remote_app="test"))

        oauth.remote_apps["test"].handle_oauth2_response = MagicMock(
            side_effect=JSONDecodeError("Expecting value", "", 0)
        )

        state = serializer.dumps(
            {
                "app": "test",
                "sid": _create_identifier(),
                "next": None,
            }
        )

        with pytest.raises(JSONDecodeError):
            client.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="test",
                    code="test",
                    state=state,
                )
            )


def test_state_token(views_fixture, monkeypatch):
    """Test state token."""
    # Mock session id
    monkeypatch.setattr(
        "invenio_oauthclient._compat._create_identifier", lambda: "1234"
    )
    monkeypatch.setattr(
        "invenio_oauthclient.views.client._create_identifier", lambda: "1234"
    )
    app = views_fixture

    with app.test_client() as client:
        # Ensure remote apps have been loaded (due to before first
        # request)
        client.get(url_for("invenio_oauthclient.login", remote_app="test"))
        mock_response(app.extensions["oauthlib.client"], "test")

        # Good state token
        state = serializer.dumps(
            {
                "app": "test",
                "sid": "1234",
                "next": None,
            }
        )
        resp = client.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="test",
                code="test",
                state=state,
            )
        )
        assert resp.status_code == 200

        outdated_serializer = TimedJSONWebSignatureSerializer(
            app.config["SECRET_KEY"],
            expires_in=0,
        )

        # Bad state - timeout
        state1 = outdated_serializer.dumps(
            {
                "app": "test",
                "sid": "1234",
                "next": None,
            }
        )
        # Bad state - app
        state2 = serializer.dumps(
            # State for another existing app (test_invalid exists)
            {
                "app": "test_invalid",
                "sid": "1234",
                "next": None,
            }
        )
        # Bad state - sid
        state3 = serializer.dumps(
            # State for another existing app (test_invalid exists)
            {
                "app": "test",
                "sid": "bad",
                "next": None,
            }
        )
        time.sleep(1)
        for s in [state1, state2, state3]:
            resp = client.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="test",
                    code="test",
                    state=s,
                )
            )
            assert resp.status_code == 403


def test_no_remote_app(views_fixture):
    """Test no remote app."""
    app = views_fixture
    with app.test_client() as client:
        resp = client.get(
            url_for("invenio_oauthclient.authorized", remote_app="invalid")
        )
        assert resp.status_code == 404

        resp = client.get(
            url_for("invenio_oauthclient.disconnect", remote_app="invalid")
        )
        assert resp.status_code == 404

        resp = client.get(url_for("invenio_oauthclient.signup", remote_app="invalid"))
        assert resp.status_code == 404


def test_token_getter_setter(views_fixture, monkeypatch):
    """Test token getter setter."""
    # Mock session id
    monkeypatch.setattr(
        "invenio_oauthclient._compat._create_identifier", lambda: "1234"
    )
    monkeypatch.setattr(
        "invenio_oauthclient.views.client._create_identifier", lambda: "1234"
    )

    app = views_fixture
    oauth = app.extensions["oauthlib.client"]

    # Mock user
    user = MagicMock()
    user.id = 1
    user.get_id = MagicMock(return_value=1)
    user.is_anonymous = False

    with app.test_client() as c:
        login_user_via_session(c, user)
        # First call login to be redirected
        res = c.get(url_for("invenio_oauthclient.login", remote_app="full"))
        assert res.status_code == 302
        assert res.location.startswith(oauth.remote_apps["full"].authorize_url)
        state = parse_qs(urlparse(res.location).query)["state"][0]

        # Mock resposen class
        mock_response(app.extensions["oauthlib.client"], "full")

        # Imitate that the user authorized our request in the remote
        # application.
        c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="full",
                code="test",
                state=state,
            )
        )

        # Assert if everything is as it should be.
        from flask import session as flask_session

        assert flask_session["oauth_token_full"] == ("test_access_token", "")

        t = RemoteToken.get(1, "fullid")
        assert t.remote_account.client_id == "fullid"
        assert t.access_token == "test_access_token"
        assert RemoteToken.query.count() == 1

        # Mock a new authorized request
        mock_response(
            app.extensions["oauthlib.client"],
            "full",
            data={
                "access_token": "new_access_token",
                "scope": "",
                "token_type": "bearer",
            },
        )

        c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="full",
                code="test",
                state=state,
            )
        )

        t = RemoteToken.get(1, "fullid")
        assert t.access_token == "new_access_token"
        assert RemoteToken.query.count() == 1

        val = token_getter(app.extensions["oauthlib.client"].remote_apps["full"])
        assert val == ("new_access_token", "")

        # Disconnect account
        res = c.get(
            url_for(
                "invenio_oauthclient.disconnect",
                remote_app="full",
            )
        )
        assert res.status_code == 302
        assert res.location.endswith(url_for("invenio_oauthclient_settings.index"))
        # Assert that remote account have been removed.
        t = RemoteToken.get(1, "fullid")
        assert t is None
        # TODO: Figure out what is leaving session open & blocked
        db.session.close()


def test_rejected(views_fixture, monkeypatch):
    """Test rejected."""
    # Mock session id
    monkeypatch.setattr(
        "invenio_oauthclient._compat._create_identifier", lambda: "1234"
    )
    monkeypatch.setattr(
        "invenio_oauthclient.views.client._create_identifier", lambda: "1234"
    )

    app = views_fixture
    oauth = app.extensions["oauthlib.client"]

    # Mock user id
    user = MagicMock()
    user.get_id = MagicMock(return_value=1)
    user.is_authenticated = MagicMock(return_value=True)

    with app.test_client() as c:
        login_user_via_session(c, user)
        # First call login to be redirected
        res = c.get(url_for("invenio_oauthclient.login", remote_app="full"))
        assert res.status_code == 302
        assert res.location.startswith(oauth.remote_apps["full"].authorize_url)

        # Mock response to imitate an invalid response. Here, an
        # example from GitHub when the code is expired.
        mock_response(
            app.extensions["oauthlib.client"],
            "full",
            data=dict(
                error_uri="http://developer.github.com/v3/oauth/"
                "#bad-verification-code",
                error_description="The code passed is " "incorrect or expired.",
                error="bad_verification_code",
            ),
        )

        # Imitate that the user authorized our request in the remote
        # application (however, the remote app will son reply with an
        # error)
        state = serializer.dumps(
            {
                "app": "full",
                "sid": "1234",
                "next": None,
            }
        )

        res = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="full",
                code="test",
                state=state,
            )
        )
        assert res.status_code == 302


def test_settings_view(views_fixture):
    """Test settings view."""
    app = views_fixture
    app.login_manager.login_view = None
    login_manager = app.login_manager
    datastore = app.extensions["invenio-accounts"].datastore

    @login_manager.user_loader
    def load_user(user_id):
        return user

    @app.route("/foo_login")
    def login():
        user = datastore.find_user(email="existing@inveniosoftware.org")
        login_user(user)
        return "Logged In"

    with app.app_context():
        with app.test_client() as client:
            user = datastore.find_user(email="existing@inveniosoftware.org")
            RemoteAccount.create(user.get_id(), "testid", None)

            resp = client.get(
                url_for("invenio_oauthclient_settings.index"), follow_redirects=False
            )
            assert resp.status_code == 401

            # make a fake login (using my login function)
            client.get("/foo_login", follow_redirects=True)

            resp = client.get(
                url_for("invenio_oauthclient_settings.index"), follow_redirects=True
            )
            assert resp.status_code == 200
            assert b"MyLinkedTestAccount" in resp.data
            assert url_for(
                "invenio_oauthclient.disconnect", remote_app="test"
            ) in resp.data.decode("utf-8")
            assert url_for(
                "invenio_oauthclient.login", remote_app="full"
            ) in resp.data.decode("utf-8")
            assert url_for(
                "invenio_oauthclient.login", remote_app="test_invalid"
            ) in resp.data.decode("utf-8")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 University of Chicago.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for globus oauth remote app."""

import json

import pytest
from flask import session, url_for
from flask_login import current_user
from flask_oauthlib.client import OAuthResponse
from flask_security import login_user
from flask_security.utils import hash_password
from helpers import mock_remote_get, mock_response
from invenio_accounts.models import User
from invenio_db import db
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient._compat import _create_identifier
from invenio_oauthclient.errors import OAuthResponseError
from invenio_oauthclient.models import RemoteAccount, RemoteToken, UserIdentity
from invenio_oauthclient.views.client import serializer


def _get_state():
    return serializer.dumps(
        {
            "app": "globus",
            "sid": _create_identifier(),
            "next": None,
        }
    )


def test_login(app):
    """Test globus login."""
    client = app.test_client()

    resp = client.get(
        url_for("invenio_oauthclient.login", remote_app="globus", next="/someurl/")
    )
    assert resp.status_code == 302

    params = parse_qs(urlparse(resp.location).query)
    assert params["response_type"], ["code"]
    assert params["scope"] == ["openid email profile"]
    assert params["redirect_uri"]
    assert params["client_id"]
    assert params["state"]


def test_authorized_signup_valid_user(app, example_globus):
    """Test authorized callback with sign-up."""

    with app.test_client() as c:
        # User login with email 'info'
        ioc = app.extensions["oauthlib.client"]

        # Ensure remote apps have been loaded (due to before first request)
        resp = c.get(url_for("invenio_oauthclient.login", remote_app="globus"))
        assert resp.status_code == 302

        example_info, example_token, example_account_id = example_globus
        mock_response(app.extensions["oauthlib.client"], "globus", example_token)
        example_info.update(example_account_id)
        oauth_resp = OAuthResponse(
            resp=None, content=json.dumps(example_info), content_type="application/json"
        )
        mock_remote_get(ioc, "globus", oauth_resp)

        # User authorized the requests and is redirect back
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="globus",
                code="test",
                state=_get_state(),
            )
        )
        assert resp.status_code == 302
        assert resp.location == "/account/settings/linkedaccounts/"

        # Assert database state (Sign-up complete)
        user = User.query.filter_by(email="carberry@inveniosoftware.org").one()
        remote = RemoteAccount.query.filter_by(user_id=user.id).one()
        RemoteToken.query.filter_by(id_remote_account=remote.id).one()
        assert user.active
        assert user.confirmed_at

        # Disconnect link
        # should not work, because it's the user's only means of login
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="globus"))
        assert resp.status_code == 400

        user = User.query.filter_by(email="carberry@inveniosoftware.org").one()
        assert (
            1
            == UserIdentity.query.filter_by(
                method="globus",
                id_user=user.id,
            ).count()
        )

        # set a password for the user
        user.password = hash_password("1234")
        db.session.commit()

        # Disconnect again
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="globus"))
        assert resp.status_code == 302

        # User exists
        user = User.query.filter_by(email="carberry@inveniosoftware.org").one()
        assert (
            0
            == UserIdentity.query.filter_by(
                method="globus",
                id_user=user.id,
            ).count()
        )
        assert RemoteAccount.query.filter_by(user_id=user.id).count() == 0
        assert RemoteToken.query.count() == 0

        # User authorized the requests and is redirect back
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="globus",
                code="test",
                state=_get_state(),
            )
        )
        assert resp.status_code == 302
        assert resp.location == "/account/settings/linkedaccounts/"

        # check that exist only one account
        user = User.query.filter_by(email="carberry@inveniosoftware.org").one()
        assert User.query.count() == 1


def test_authorized_reject(app):
    """Test a rejected request."""
    with app.test_client() as c:
        c.get(url_for("invenio_oauthclient.login", remote_app="globus"))
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="globus",
                error="access_denied",
                error_description="User denied access",
                state=_get_state(),
            )
        )
        assert resp.status_code in (301, 302)
        assert resp.location == "/"
        # Check message flash
        assert session["_flashes"][0][0] == "info"


def test_authorized_already_authenticated(app, models_fixture, example_globus):
    """Test authorized callback with sign-up."""
    datastore = app.extensions["invenio-accounts"].datastore
    login_manager = app.login_manager

    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    @login_manager.user_loader
    def load_user(user_id):
        return user

    @app.route("/foo_login")
    def login():
        login_user(user)
        return "Logged In"

    with app.test_client() as client:
        # make a fake login (using my login function)
        client.get("/foo_login", follow_redirects=True)
        # Ensure remote apps have been loaded (due to before first request)
        client.get(url_for("invenio_oauthclient.login", remote_app="globus"))

        ioc = app.extensions["oauthlib.client"]
        example_info, example_token, example_account_id = example_globus
        mock_response(app.extensions["oauthlib.client"], "globus", example_token)
        example_info.update(example_account_id)
        oauth_resp = OAuthResponse(
            resp=None, content=json.dumps(example_info), content_type="application/json"
        )
        mock_remote_get(ioc, "globus", oauth_resp)

        # User then goes to 'Linked accounts' and clicks 'Connect'
        resp = client.get(
            url_for("invenio_oauthclient.login", remote_app="globus", next="/someurl/")
        )
        assert resp.status_code == 302

        # User authorized the requests and is redirected back
        resp = client.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="globus",
                code="test",
                state=_get_state(),
            )
        )

        # Assert database state (Sign-up complete)
        u = User.query.filter_by(email=existing_email).one()
        remote = RemoteAccount.query.filter_by(user_id=u.id).one()
        RemoteToken.query.filter_by(id_remote_account=remote.id).one()

        # Disconnect link
        resp = client.get(
            url_for("invenio_oauthclient.disconnect", remote_app="globus")
        )
        assert resp.status_code == 302

        # User exists
        u = User.query.filter_by(email=existing_email).one()
        assert (
            0
            == UserIdentity.query.filter_by(
                method="globus", id_user=u.id, id="globususer"
            ).count()
        )
        assert RemoteAccount.query.filter_by(user_id=u.id).count() == 0
        assert RemoteToken.query.count() == 0


def test_not_authenticated(app):
    """Test disconnect when user is not authenticated."""
    with app.test_client() as client:
        assert not current_user.is_authenticated
        resp = client.get(
            url_for("invenio_oauthclient.disconnect", remote_app="globus")
        )
        assert resp.status_code == 302


def test_bad_provider_response(app, example_globus):
    with app.test_client() as c:

        class MockResponse:
            code = 403

        # User login with email 'info'
        ioc = app.extensions["oauthlib.client"]

        # Ensure remote apps have been loaded (due to before first request)
        resp = c.get(url_for("invenio_oauthclient.login", remote_app="globus"))
        assert resp.status_code == 302

        _, example_token, _ = example_globus
        mock_response(app.extensions["oauthlib.client"], "globus", example_token)
        oauth_resp = OAuthResponse(
            resp=MockResponse(), content=None, content_type="application/json"
        )
        mock_remote_get(ioc, "globus", oauth_resp)

        with pytest.raises(OAuthResponseError):
            c.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="globus",
                    code="test",
                    state=_get_state(),
                )
            )


def test_invalid_user_id_response(app, example_globus):
    with app.test_client() as c:
        # User login with email 'info'
        ioc = app.extensions["oauthlib.client"]

        # Ensure remote apps have been loaded (due to before first request)
        resp = c.get(url_for("invenio_oauthclient.login", remote_app="globus"))
        assert resp.status_code == 302

        example_info, example_token, _ = example_globus
        mock_response(app.extensions["oauthlib.client"], "globus", example_token)
        oauth_resp = OAuthResponse(
            resp=None, content=json.dumps(example_info), content_type="application/json"
        )
        mock_remote_get(ioc, "globus", oauth_resp)

        with pytest.raises(OAuthResponseError):
            c.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="globus",
                    code="test",
                    state=_get_state(),
                )
            )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for ORCID oauth remote app."""

import httpretty
from flask import session, url_for
from flask_login import current_user
from flask_security import login_user
from flask_security.utils import hash_password
from helpers import check_response_redirect_url_args, get_state, mock_response
from invenio_accounts.models import User
from invenio_db import db
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient.contrib.orcid import account_info
from invenio_oauthclient.handlers import token_session_key
from invenio_oauthclient.models import RemoteAccount, RemoteToken, UserIdentity


def test_account_info(app_rest, example_orcid):
    """Test account info extraction."""
    client = app_rest.test_client()
    ioc = app_rest.extensions["oauthlib.client"]
    # Ensure remote apps have been loaded (due to before first
    # request)
    client.get(url_for("invenio_oauthclient.rest_login", remote_app="orcid"))

    example_data, example_account_info = example_orcid

    assert account_info(ioc.remote_apps["orcid"], example_data) == example_account_info

    assert account_info(ioc.remote_apps["orcid"], {}) == dict(
        external_id=None,
        external_method="orcid",
        user=dict(profile=dict(full_name=None)),
    )


def test_login(app_rest, example_orcid):
    """Test ORCID login."""
    client = app_rest.test_client()

    resp = client.get(
        url_for("invenio_oauthclient.rest_login", remote_app="orcid", next="/someurl/")
    )
    assert resp.status_code == 302

    params = parse_qs(urlparse(resp.location).query)
    assert params["response_type"], ["code"]
    assert params["show_login"] == ["true"]
    assert params["scope"] == ["/authenticate"]
    assert params["redirect_uri"]
    assert params["client_id"]
    assert params["state"]


def test_authorized_signup(app_rest_with_userprofiles, example_orcid, orcid_bio):
    """Test authorized callback with sign-up."""
    app = app_rest_with_userprofiles
    example_data, example_account_info = example_orcid
    example_email = "orcidtest@inveniosoftware.org"

    with app.test_client() as c:
        # Ensure remote apps have been loaded (due to before first
        # request)
        c.get(url_for("invenio_oauthclient.rest_login", remote_app="orcid"))

        mock_response(app.extensions["oauthlib.client"], "orcid", example_data)

        # User authorized the requests and is redirect back
        resp = c.get(
            url_for(
                "invenio_oauthclient.rest_authorized",
                remote_app="orcid",
                code="test",
                state=get_state("orcid"),
            )
        )
        assert resp.status_code == 302
        assert resp.location == (
            url_for("invenio_oauthclient.rest_signup", remote_app="orcid")
        )

        # User load sign-up page.
        resp = c.get(url_for("invenio_oauthclient.rest_signup", remote_app="orcid"))
        assert resp.status_code == 302

        account_info = session[token_session_key("orcid") + "_account_info"]
        data = {
            "email": example_email,
            "password": "123456",
            "profile.username": "pippo",
            "profile.full_name": account_info["user"]["profile"]["full_name"],
            "profile.affiliations": "CERN",
        }

        # Mock request to ORCID to get user bio.
        httpretty.enable()
        httpretty.register_uri(
            httpretty.GET,
            "http://orcid.org/{0}/orcid-bio".format(example_data["orcid"]),
            body=orcid_bio,
            content_type="application/orcid+json; qs=2;charset=UTF-8",
        )

        # User fills form to register
        resp = c.post(
            url_for("invenio_oauthclient.rest_signup", remote_app="orcid"),
            data=data,
        )
        assert resp.status_code == 200
        httpretty.disable()

        # Assert database state (Sign-up complete)
        user = User.query.filter_by(email=example_email).one()
        UserIdentity.query.filter_by(
            method="orcid", id_user=user.id, id=example_data["orcid"]
        ).one()
        assert user.user_profile["full_name"] == "Josiah Carberry"
        #  assert user.given_names == 'Josiah'
        #  assert user.family_name == 'Carberry'
        # check that the user's email is not yet validated
        assert user.active
        assert not user.confirmed_at
        # check that the validation email has been sent
        #  assert hasattr(locmem, 'outbox') and len(locmem.outbox) == 1

        # Disconnect link
        # should not work, because it's the user's only means of login
        resp = c.get(url_for("invenio_oauthclient.rest_disconnect", remote_app="orcid"))
        assert resp.status_code == 400

        user = User.query.filter_by(email=example_email).one()
        assert (
            1
            == UserIdentity.query.filter_by(
                method="orcid", id_user=user.id, id=example_data["orcid"]
            ).count()
        )

        # set a password for the user
        user.password = hash_password("1234")
        db.session.commit()

        # Disconnect again
        resp = c.get(url_for("invenio_oauthclient.rest_disconnect", remote_app="orcid"))
        assert resp.status_code == 302

        # User exists
        user = User.query.filter_by(email=example_email).one()
        # UserIdentity removed.
        assert (
            0
            == UserIdentity.query.filter_by(
                method="orcid", id_user=user.id, id=example_data["orcid"]
            ).count()
        )

        assert RemoteAccount.query.filter_by(user_id=user.id).count() == 0
        assert RemoteToken.query.count() == 0


def test_authorized_reject(app_rest, example_orcid):
    """Test a rejected request."""
    with app_rest.test_client() as c:
        c.get(url_for("invenio_oauthclient.rest_login", remote_app="orcid"))
        resp = c.get(
            url_for(
                "invenio_oauthclient.rest_authorized",
                remote_app="orcid",
                error="access_denied",
                error_description="User denied access",
                state=get_state("orcid"),
            )
        )
        assert resp.status_code in (301, 302)
        expected_url_args = {
            "message": "You rejected the authentication request.",
            "code": 400,
        }
        check_response_redirect_url_args(resp, expected_url_args)


def test_authorized_already_authenticated(
    app_rest, models_fixture, example_orcid, orcid_bio
):
    """Test authorized callback with sign-up."""
    datastore = app_rest.extensions["invenio-accounts"].datastore
    login_manager = app_rest.login_manager

    example_data, example_account_info = example_orcid
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    @login_manager.user_loader
    def load_user(user_id):
        return user

    @app_rest.route("/foo_login")
    def login():
        login_user(user)
        return "Logged In"

    with app_rest.test_client() as client:
        # make a fake login (using my login function)
        client.get("/foo_login", follow_redirects=True)

        # Ensure remote apps have been loaded (due to before first
        # request)
        client.get(url_for("invenio_oauthclient.rest_login", remote_app="orcid"))

        # Mock access token request
        mock_response(app_rest.extensions["oauthlib.client"], "orcid", example_data)

        # Mock request to ORCID to get user bio.
        httpretty.enable()
        httpretty.register_uri(
            httpretty.GET,
            "https://pub.orcid.org/v1.2/{0}/orcid-bio".format(example_data["orcid"]),
            body=orcid_bio,
            content_type="application/orcid+json; qs=2;charset=UTF-8",
        )

        # User then goes to 'Linked accounts' and clicks 'Connect'
        resp = client.get(
            url_for(
                "invenio_oauthclient.rest_login", remote_app="orcid", next="/someurl/"
            )
        )
        assert resp.status_code == 302

        # User authorized the requests and is redirected back
        resp = client.get(
            url_for(
                "invenio_oauthclient.rest_authorized",
                remote_app="orcid",
                code="test",
                state=get_state("orcid"),
            )
        )
        httpretty.disable()

        # Assert database state (Sign-up complete)
        u = User.query.filter_by(email=existing_email).one()
        UserIdentity.query.filter_by(
            method="orcid", id_user=u.id, id=example_data["orcid"]
        ).one()
        # FIXME see contrib/orcid.py line 167
        # assert u.given_names == 'Josiah'
        # assert u.family_name == 'Carberry'

        # Disconnect link
        resp = client.get(
            url_for("invenio_oauthclient.rest_disconnect", remote_app="orcid")
        )
        assert resp.status_code == 302

        # User exists
        u = User.query.filter_by(email=existing_email).one()
        # UserIdentity removed.
        assert (
            0
            == UserIdentity.query.filter_by(
                method="orcid", id_user=u.id, id=example_data["orcid"]
            ).count()
        )


def test_not_authenticated(app_rest, models_fixture):
    """Test disconnect when user is not authenticated."""
    with app_rest.test_client() as client:
        assert not current_user.is_authenticated
        resp = client.get(
            url_for("invenio_oauthclient.rest_disconnect", remote_app="orcid")
        )
        assert resp.status_code == 302



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create oauthclient tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = "97bbc733896c"
down_revision = "44ab9963e8cf"
branch_labels = ()
depends_on = "9848d0149abd"


def upgrade():
    """Upgrade database."""
    op.create_table(
        "oauthclient_remoteaccount",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("client_id", sa.String(length=255), nullable=False),
        sa.Column("extra_data", sqlalchemy_utils.JSONType(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "client_id"),
    )
    op.create_table(
        "oauthclient_useridentity",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("method", sa.String(length=255), nullable=False),
        sa.Column("id_user", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id_user"],
            ["accounts_user.id"],
        ),
        sa.PrimaryKeyConstraint("id", "method"),
    )
    op.create_index(
        "useridentity_id_user_method",
        "oauthclient_useridentity",
        ["id_user", "method"],
        unique=True,
    )
    op.create_table(
        "oauthclient_remotetoken",
        sa.Column("id_remote_account", sa.Integer(), nullable=False),
        sa.Column("token_type", sa.String(length=40), nullable=False),
        sa.Column("access_token", sqlalchemy_utils.EncryptedType(), nullable=False),
        sa.Column("secret", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id_remote_account"],
            ["oauthclient_remoteaccount.id"],
            name="fk_oauthclient_remote_token_remote_account",
        ),
        sa.PrimaryKeyConstraint("id_remote_account", "token_type"),
    )


def downgrade():
    """Downgrade database."""
    ctx = op.get_context()
    insp = Inspector.from_engine(ctx.connection.engine)

    op.drop_table("oauthclient_remotetoken")

    for fk in insp.get_foreign_keys("oauthclient_useridentity"):
        if fk["referred_table"] == "accounts_user":
            op.drop_constraint(
                op.f(fk["name"]), "oauthclient_useridentity", type_="foreignkey"
            )
    op.drop_index("useridentity_id_user_method", table_name="oauthclient_useridentity")
    op.drop_table("oauthclient_useridentity")
    op.drop_table("oauthclient_remoteaccount")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 TU Wien.
#
# Invenio-Keycloak is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuthClient endpoint handlers for communication with Keycloak.

The handler functions provided in this module are tailored to communicate
with Keycloak using OpenID-Connect.
To use them, they must be referenced in a REMOTE_APP configuration dictionary,
e.g.:

.. code-block:: python

    KEYCLOAK_REMOTE_APP = {
        # ...

        "authorized_handler": "invenio_oauthclient.handlers"
                              ":authorized_signup_handler",
        "disconnect_handler": "invenio_keycloak.handlers"
                              ":disconnect_handler",
        "signup_handler": {
            "info": "invenio_keycloak.handlers:info_handler",
            "setup": "invenio_keycloak.handlers:setup_handler",
            "view": "invenio_oauthclient.handlers:signup_handler"
        },

        # ...
    }

"""

from flask import current_app, redirect, url_for
from flask_login import current_user
from invenio_db import db

from invenio_oauthclient import current_oauthclient
from invenio_oauthclient.handlers.rest import response_handler
from invenio_oauthclient.handlers.utils import require_more_than_one_external_account
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import oauth_link_external_id, oauth_unlink_external_id

from .helpers import get_user_info


def info_serializer_handler(remote, resp, token_user_info, user_info=None, **kwargs):
    """Serialize the account info response object.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :param token_user_info: The content of the authorization token response.
    :param user_info: The response of the `user info` endpoint.
    :returns: A dictionary with serialized user information.
    """
    # fill out the information required by
    # 'invenio-accounts' and 'invenio-userprofiles'.
    #
    # note: "external_id": `preferred_username` should also work,
    #       as it is seemingly not editable in Keycloak

    user_info = user_info or {}  # prevent errors when accessing None.get(...)

    email = token_user_info.get("email") or user_info["email"]
    full_name = token_user_info.get("name") or user_info.get("name")
    username = token_user_info.get("preferred_username") or user_info.get(
        "preferred_username"
    )

    return {
        "user": {
            "active": True,
            "email": email,
            "profile": {
                "full_name": full_name,
                "username": username,
            },
        },
        "external_id": token_user_info["sub"],
        "external_method": remote.name,
    }


def info_handler(remote, resp):
    """Retrieve remote account information for finding matching local users.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with the user information.
    """
    token_user_info, user_info = get_user_info(remote, resp)

    handlers = current_oauthclient.signup_handlers[remote.name]
    # `remote` param automatically injected via `make_handler` helper
    return handlers["info_serializer"](resp, token_user_info, user_info)


def setup_handler(remote, token, resp):
    """Perform additional setup after the user has been logged in."""
    token_user_info, _ = get_user_info(remote, resp)

    with db.session.begin_nested():
        # fetch the user's Keycloak ID and set it in extra_data
        keycloak_id = token_user_info["sub"]
        token.remote_account.extra_data = {
            "keycloak_id": keycloak_id,
        }

        user = token.remote_account.user
        external_id = {"id": keycloak_id, "method": remote.name}

        # link account with external Keycloak ID
        oauth_link_external_id(user, external_id)


@require_more_than_one_external_account
def _disconnect(remote, *args, **kwargs):
    """Common logic for handling disconnection of remote accounts."""
    if not current_user.is_authenticated:
        return current_app.login_manager.unauthorized()

    account = RemoteAccount.get(
        user_id=current_user.get_id(), client_id=remote.consumer_key
    )

    keycloak_id = account.extra_data.get("keycloak_id")

    if keycloak_id:
        external_id = {"id": keycloak_id, "method": remote.name}

        oauth_unlink_external_id(external_id)

    if account:
        with db.session.begin_nested():
            account.delete()


def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of the remote account."""
    _disconnect(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


def disconnect_rest_handler(remote, *args, **kwargs):
    """Handle unlinking of the remote account."""
    _disconnect(remote, *args, **kwargs)
    rconfig = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name]
    redirect_url = rconfig["disconnect_redirect_url"]
    return response_handler(remote, redirect_url)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 TU Wien.
#
# Invenio-Keycloak is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Toolkit for creating remote apps that enable sign in/up with Keycloak.

**Note:** In contrast to e.g. GitHub, Keycloak is a self-hosted solution that
requires a bit of configuration to allow OAuth client applications, such as
Invenio. An explanation of how to set up Keycloak is out of scope for this
document, where we focus on configuring Invenio.

0. Decide on your keycloak app name, for example ``keycloak``. It will be
   needed in multiple places in this configuration.
   Choose a name that has only letters and does not contain spaces. This name
   will be also used as URL path and to look for configuration variables
   such as ``OAUTHCLIENT_<APP_NAME>_REALM_URL``.

1. Set up a Keycloak server and make sure it is configured appropriately,
   i.e. a client application for Invenio is configured in a realm in Keycloak.
   In your server, set the redirect URI as
   "https://myinveniohost.com/oauth/authorized/keycloak/".
   Make sure that the the name of the app, ``keycloak``, matches the path in
   the URL.

2. Add the following items to your configuration (``invenio.cfg``).
   The ``KeycloakSettingsHelper`` class can be used to help with setting up
   the configuration values:

   .. code-block:: python

        from invenio_oauthclient.contrib import keycloak as k

        helper = k.KeycloakSettingsHelper(
            title="My organization",
            description="My organization authentication provider",
            base_url="http://yourkeycloakserver.com:8080",
            realm="invenio"
        )

        # create the configuration for Keycloak
        # because the URLs usually follow a certain schema, the settings helper
        # can be used to more easily build the configuration values:
        OAUTHCLIENT_KEYCLOAK_REALM_URL = helper.realm_url
        OAUTHCLIENT_KEYCLOAK_USER_INFO_URL = helper.user_info_url

        # Keycloak uses JWTs (https://jwt.io/) for their tokens, which
        # contain information about the target audience (AUD)
        # verification of the expected AUD value can be configured with:
        OAUTHCLIENT_KEYCLOAK_VERIFY_AUD = True
        OAUTHCLIENT_KEYCLOAK_AUD = "invenio"

        # enable/disable checking if the JWT signature has expired
        OAUTHCLIENT_KEYCLOAK_VERIFY_EXP = True

        # add Keycloak to the dictionary of remote apps
        OAUTHCLIENT_REMOTE_APPS = dict(
            keycloak=helper.remote_app,
            # ...
        )

        # set the following configuration to True to automatically use the
        # user's email address as account email
        USERPROFILES_EXTEND_SECURITY_FORMS = True

   By default, the title will be displayed as label for the login button,
    for example ``Login with My organization``. The description will be
    displayed in the user account section.

3. Grab the *Client ID* and *Client Secret* from the client application in
   Keycloak and add them to your instance configuration (``invenio.cfg``):

   .. code-block:: python

        KEYCLOAK_APP_CREDENTIALS = dict(
            consumer_key='<CLIENT ID>',
            consumer_secret='<CLIENT SECRET>',
        )

4. Now go to ``CFG_SITE_SECURE_URL/oauth/login/keycloak/`` (e.g.
   https://localhost:5000/oauth/login/keycloak/) and log in.

5. After authenticating successfully, you should see Keycloak listed under
   Linked accounts: https://localhost:5000/account/settings/linkedaccounts/
"""

from .handlers import (
    disconnect_handler,
    disconnect_rest_handler,
    info_handler,
    setup_handler,
)
from .settings import KeycloakSettingsHelper

__all__ = (
    "disconnect_handler",
    "disconnect_rest_handler",
    "info_handler",
    "setup_handler",
    "KeycloakSettingsHelper",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Client blueprint used to handle OAuth callbacks."""

from flask import Blueprint, abort, current_app, redirect, request, session, url_for
from flask_oauthlib.client import OAuthException
from invenio_accounts.views import login as base_login
from invenio_db import db
from itsdangerous import BadData

from .._compat import _create_identifier
from ..errors import OAuthRemoteNotFound
from ..handlers import set_session_next_url
from ..handlers.rest import response_handler
from ..proxies import current_oauthclient
from ..utils import get_safe_redirect_target, serializer

blueprint = Blueprint(
    "invenio_oauthclient",
    __name__,
    url_prefix="/oauth",
    static_folder="../static",
    template_folder="../templates",
)


rest_blueprint = Blueprint(
    "invenio_oauthclient",
    __name__,
    url_prefix="/oauth",
    static_folder="../static",
    template_folder="../templates",
)


@blueprint.record_once
def post_ext_init(state):
    """Setup blueprint."""
    app = state.app

    app.config.setdefault(
        "OAUTHCLIENT_SITENAME", app.config.get("THEME_SITENAME", "Invenio")
    )
    app.config.setdefault(
        "OAUTHCLIENT_BASE_TEMPLATE",
        app.config.get("BASE_TEMPLATE", "invenio_oauthclient/base.html"),
    )
    app.config.setdefault(
        "OAUTHCLIENT_COVER_TEMPLATE",
        app.config.get("COVER_TEMPLATE", "invenio_oauthclient/base_cover.html"),
    )
    app.config.setdefault(
        "OAUTHCLIENT_SETTINGS_TEMPLATE",
        app.config.get("SETTINGS_TEMPLATE", "invenio_oauthclient/settings/base.html"),
    )


@blueprint.route("/login")
def auto_redirect_login(*args, **kwargs):
    """Handles automatic redirect to external auth service.

    The login endpoint will redirect automatically to the external
    auth service is the following conditions are met:

    * local login is disabled
    * redirect to external login is enabled
    * only one external auth service is configured

    This function should be set as value of the invenio-accounts
    config var ``ACCOUNTS_LOGIN_VIEW_FUNCTION``. It should be defined in
    the Invenio application configuration to ensure that is correctly loaded.
    """
    local_login_enabled = current_app.config.get("ACCOUNTS_LOCAL_LOGIN_ENABLED", False)
    auto_redirect_enabled = current_app.config.get(
        "OAUTHCLIENT_AUTO_REDIRECT_TO_EXTERNAL_LOGIN", False
    )
    would_redirect = auto_redirect_enabled and not local_login_enabled
    remote_apps = list(current_oauthclient.oauth.remote_apps)

    if would_redirect and len(remote_apps) == 1:
        # if local login is disabled and we only have one OAuth2 remote app
        # configured, we forward directly to that
        url = url_for("invenio_oauthclient.login", remote_app=remote_apps[0])
        return redirect(url)

    else:
        return base_login(*args, **kwargs)


def _login(remote_app, authorized_view_name):
    """Send user to remote application for authentication."""
    oauth = current_oauthclient.oauth
    if remote_app not in oauth.remote_apps:
        raise OAuthRemoteNotFound()

    # Get redirect target in safe manner.
    next_param = get_safe_redirect_target(arg="next")

    # Redirect URI - must be registered in the remote service.
    callback_url = url_for(
        authorized_view_name, remote_app=remote_app, _external=True, _scheme="https"
    )

    # Create a JSON Web Token that expires after OAUTHCLIENT_STATE_EXPIRES
    # seconds.
    state_token = serializer.dumps(
        {
            "app": remote_app,
            "next": next_param,
            "sid": _create_identifier(),
        }
    )
    return oauth.remote_apps[remote_app].authorize(
        callback=callback_url,
        state=state_token,
    )


@blueprint.route("/login/<remote_app>/")
def login(remote_app):
    """Send user to remote application for authentication."""
    try:
        return _login(remote_app, ".authorized")
    except OAuthRemoteNotFound:
        return abort(404)


@rest_blueprint.route("/login/<remote_app>/")
def rest_login(remote_app):
    """Send user to remote application for authentication."""
    try:
        return _login(remote_app, ".rest_authorized")
    except OAuthRemoteNotFound:
        abort(404)


def _authorized(remote_app=None):
    """Authorized handler callback."""
    if remote_app not in current_oauthclient.handlers:
        return abort(404)

    state_token = request.args.get("state")

    # Verify state parameter
    assert state_token
    # Checks authenticity and integrity of state and decodes the value.
    state = serializer.loads(state_token)
    # Verify that state is for this session, app and that next parameter
    # have not been modified.
    assert state["sid"] == _create_identifier()
    assert state["app"] == remote_app
    # Store next URL
    set_session_next_url(remote_app, state["next"])

    handler = current_oauthclient.handlers[remote_app]()
    return handler


@blueprint.route("/authorized/<remote_app>/")
def authorized(remote_app=None):
    """Authorized handler callback."""
    try:
        return _authorized(remote_app)
    except OAuthRemoteNotFound:
        return abort(404)
    except (AssertionError, BadData):
        if current_app.config.get("OAUTHCLIENT_STATE_ENABLED", True) or (
            not (current_app.debug or current_app.testing)
        ):
            abort(403)
    except OAuthException as e:
        if e.type == "invalid_response":
            current_app.logger.warning(
                "{message} ({data})".format(message=e.message, data=e.data)
            )
            abort(500)
        else:
            raise


@rest_blueprint.route("/authorized/<remote_app>/")
def rest_authorized(remote_app=None):
    """Authorized handler callback."""
    try:
        return _authorized(remote_app)
    except OAuthRemoteNotFound:
        abort(404)
    except (AssertionError, BadData) as e:
        current_app.logger.error(str(e))
        if current_app.config.get("OAUTHCLIENT_STATE_ENABLED", True) or (
            not (current_app.debug or current_app.testing)
        ):
            return response_handler(
                None,
                current_app.config["OAUTHCLIENT_REST_DEFAULT_ERROR_REDIRECT_URL"],
                payload=dict(message="Invalid state.", code=403),
            )
    except OAuthException as e:
        current_app.logger.error(str(e))
        if e.type == "invalid_response":
            return response_handler(
                None,
                current_app.config["OAUTHCLIENT_REST_DEFAULT_ERROR_REDIRECT_URL"],
                payload=dict(message="Invalid response.", code=500),
            )
        else:
            raise


def _signup(remote_app):
    """Extra signup step."""
    if remote_app not in current_oauthclient.signup_handlers:
        raise OAuthRemoteNotFound()
    return current_oauthclient.signup_handlers[remote_app]["view"]()


@blueprint.route("/signup/<remote_app>/", methods=["GET", "POST"])
def signup(remote_app):
    """Extra signup step."""
    try:
        res = _signup(remote_app)
        return abort(404) if res is None else res
    except OAuthRemoteNotFound:
        return abort(404)


@rest_blueprint.route("/signup/<remote_app>/", methods=["GET", "POST"])
def rest_signup(remote_app):
    """Extra signup step."""
    try:
        res = _signup(remote_app)
        return abort(404) if res is None else res
    except OAuthRemoteNotFound:
        abort(404)


def _disconnect(remote_app):
    """Extra signup step."""
    if remote_app not in current_oauthclient.signup_handlers:
        raise OAuthRemoteNotFound()
    ret = current_oauthclient.disconnect_handlers[remote_app]()
    db.session.commit()
    return ret


@blueprint.route("/disconnect/<remote_app>/")
def disconnect(remote_app):
    """Disconnect user from remote application.

    Removes application as well as associated information.
    """
    try:
        return _disconnect(remote_app)
    except OAuthRemoteNotFound:
        abort(404)


@rest_blueprint.route("/disconnect/<remote_app>/")
def rest_disconnect(remote_app):
    """Disconnect user from remote application.

    Removes application as well as associated information.
    """
    try:
        return _disconnect(remote_app)
    except OAuthRemoteNotFound:
        abort(404)


@blueprint.route("/logout")
def post_logout():
    """Client logout view.

    This URL should be called by setting `SECURITY_POST_LOGOUT_VIEW = /oauth/logout`
    """
    remote_name = session.pop("OAUTHCLIENT_SESSION_REMOTE_NAME", None)
    if remote_name:
        logout_url = current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote_name].get(
            "logout_url"
        )
        if logout_url:
            return redirect(logout_url, code=302)

    return redirect("/")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2018 University of Chicago.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import json
import os
import shutil
import tempfile

import pytest
from flask import Flask
from flask_mail import Mail
from flask_menu import Menu as FlaskMenu
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB, db
from invenio_i18n import Babel, InvenioI18N
from invenio_userprofiles import InvenioUserProfiles
from invenio_userprofiles.views import blueprint_ui_init
from sqlalchemy_utils.functions import create_database, database_exists, drop_database

from invenio_oauthclient import InvenioOAuthClient, InvenioOAuthClientREST
from invenio_oauthclient.contrib.cern_openid import REMOTE_APP as CERN_OPENID_REMOTE_APP
from invenio_oauthclient.contrib.cern_openid import (
    REMOTE_REST_APP as CERN_OPENID_REMOTE_REST_APP,
)
from invenio_oauthclient.contrib.github import REMOTE_APP as GITHUB_REMOTE_APP
from invenio_oauthclient.contrib.github import REMOTE_REST_APP as GITHUB_REMOTE_REST_APP
from invenio_oauthclient.contrib.globus import REMOTE_APP as GLOBUS_REMOTE_APP
from invenio_oauthclient.contrib.globus import REMOTE_REST_APP as GLOBUS_REMOTE_REST_APP
from invenio_oauthclient.contrib.keycloak import KeycloakSettingsHelper
from invenio_oauthclient.contrib.orcid import REMOTE_APP as ORCID_REMOTE_APP
from invenio_oauthclient.contrib.orcid import REMOTE_REST_APP as ORCID_REMOTE_REST_APP
from invenio_oauthclient.utils import _create_registrationform
from invenio_oauthclient.views.client import blueprint as blueprint_client
from invenio_oauthclient.views.client import rest_blueprint
from invenio_oauthclient.views.settings import blueprint as blueprint_settings

from invenio_oauthclient._compat import monkey_patch_werkzeug  # noqa isort:skip

try:
    from werkzeug.middleware.dispatcher import DispatcherMiddleware
except ImportError:
    from werkzeug.wsgi import DispatcherMiddleware

from flask_oauthlib.client import OAuth as FlaskOAuth  # noqa isort:skip
from flask_oauthlib.client import OAuthResponse  # noqa isort:skip


@pytest.fixture
def base_app(request):
    """Flask application fixture without OAuthClient initialized."""
    # allow HTTP for keycloak tests, and create the KEYCLOAK_REMOTE_APP
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
    base_url, realm = "http://localhost:8080", "test"
    helper = KeycloakSettingsHelper(
        title="Keycloak", description="", base_url=base_url, realm=realm
    )
    KEYCLOAK_REMOTE_APP = helper.remote_app

    instance_path = tempfile.mkdtemp()
    base_app = Flask("testapp")
    base_app.config.update(
        ACCOUNTS_LOCAL_LOGIN_ENABLED=True,
        TESTING=True,
        WTF_CSRF_ENABLED=False,
        LOGIN_DISABLED=False,
        CACHE_TYPE="simple",
        OAUTHCLIENT_SIGNUP_FORM=_create_registrationform,
        OAUTHCLIENT_REMOTE_APPS=dict(
            cern_openid=CERN_OPENID_REMOTE_APP,
            orcid=ORCID_REMOTE_APP,
            github=GITHUB_REMOTE_APP,
            globus=GLOBUS_REMOTE_APP,
            keycloak=KEYCLOAK_REMOTE_APP,
        ),
        OAUTHCLIENT_REST_REMOTE_APPS=dict(
            cern_openid=CERN_OPENID_REMOTE_REST_APP,
            orcid=ORCID_REMOTE_REST_APP,
            github=GITHUB_REMOTE_REST_APP,
            globus=GLOBUS_REMOTE_REST_APP,
        ),
        OAUTHCLIENT_STATE_EXPIRES=300,
        GITHUB_APP_CREDENTIALS=dict(
            consumer_key="github_key_changeme",
            consumer_secret="github_secret_changeme",
        ),
        ORCID_APP_CREDENTIALS=dict(
            consumer_key="orcid_key_changeme",
            consumer_secret="orcid_secret_changeme",
        ),
        CERN_APP_OPENID_CREDENTIALS=dict(
            consumer_key="cern_key_changeme",
            consumer_secret="cern_secret_changeme",
        ),
        GLOBUS_APP_CREDENTIALS=dict(
            consumer_key="globus_key_changeme",
            consumer_secret="globus_secret_changeme",
        ),
        TEST_APP_CREDENTIALS=dict(
            consumer_key="test_key_changeme",
            consumer_secret="test_secret_changeme",
        ),
        OAUTHCLIENT_KEYCLOAK_USER_INFO_URL=helper.user_info_url,
        OAUTHCLIENT_KEYCLOAK_REALM_URL=helper.realm_url,
        OAUTHCLIENT_KEYCLOAK_VERIFY_AUD=True,
        OAUTHCLIENT_KEYCLOAK_VERIFY_EXP=False,
        OAUTHCLIENT_KEYCLOAK_AUD="invenio",
        KEYCLOAK_APP_CREDENTIALS=dict(
            consumer_key="keycloak_key_changeme",
            consumer_secret="keycloak_secret_changeme",
        ),
        # use local memory mailbox
        EMAIL_BACKEND="flask_email.backends.locmem.Mail",
        SQLALCHEMY_DATABASE_URI=os.getenv("SQLALCHEMY_DATABASE_URI", "sqlite://"),
        # Alembic runs all migrations in a single transaction, and in
        # oauthclient the useridentity table can be in state where it's created
        # and then deleted in the same transaction. On PostgreSQL which
        # supports transactional DDL therefore fails if we don't run each
        # migration in its own migration.
        ALEMBIC_CONTEXT={
            "transaction_per_migration": True,
        },
        SERVER_NAME="localhost",
        DEBUG=False,
        SECRET_KEY="TEST",
        SECURITY_DEPRECATED_PASSWORD_SCHEMES=[],
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        SECURITY_PASSWORD_HASH="plaintext",
        SECURITY_PASSWORD_SCHEMES=["plaintext"],
        SECURITY_PASSWORD_SINGLE_HASH=None,
        APP_ALLOWED_HOSTS=["localhost"],
        APP_THEME=["semantic-ui"],
        THEME_ICONS={"semantic-ui": dict(link="linkify icon")},
    )
    FlaskMenu(base_app)
    Babel(base_app)
    Mail(base_app)
    InvenioDB(base_app)
    InvenioAccounts(base_app)
    InvenioI18N(base_app)

    with base_app.app_context():
        if str(db.engine.url) != "sqlite://" and not database_exists(
            str(db.engine.url)
        ):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with base_app.app_context():
            db.session.close()
            if str(db.engine.url) != "sqlite://":
                drop_database(str(db.engine.url))
            shutil.rmtree(instance_path)
            db.engine.dispose()

    request.addfinalizer(teardown)

    base_app.test_request_context().push()

    return base_app


def _init_app(app_):
    """Init OAuth app."""
    app_.config.update(
        WTF_CSRF_ENABLED=False,
    )
    FlaskOAuth(app_)
    InvenioOAuthClient(app_)
    app_.register_blueprint(blueprint_client)
    app_.register_blueprint(blueprint_settings)
    return app_


def _init_app_rest(app_):
    """Init OAuth rest app."""
    FlaskOAuth(app_)
    InvenioOAuthClientREST(app_)
    app_.register_blueprint(rest_blueprint)
    return app_


@pytest.fixture
def app(base_app):
    """Flask application fixture."""
    return _init_app(base_app)


@pytest.fixture
def app_rest(base_app, views_fixture_rest):
    """Flask application fixture."""
    return _init_app_rest(base_app)


@pytest.fixture
def app_with_csrf(base_app):
    """Flask application fixture with CSRF enabled."""
    app_ = _init_app(base_app)
    app_.config.update(
        WTF_CSRF_ENABLED=True,
    )
    return app_


@pytest.fixture
def app_rest_with_userprofiles(app_rest):
    """Configure userprofiles module with CSRF disabled."""
    app_rest.config.update(
        USERPROFILES_EXTEND_SECURITY_FORMS=True,
        WTF_CSRF_ENABLED=False,
    )
    return _init_userprofiles(app_rest)


def _init_userprofiles(app_):
    """Init userprofiles module."""
    InvenioUserProfiles(app_)
    app_.register_blueprint(blueprint_ui_init)
    return app_


@pytest.fixture
def app_with_userprofiles(app):
    """Configure userprofiles module with CSRF disabled."""
    app.config.update(
        USERPROFILES_EXTEND_SECURITY_FORMS=True,
        WTF_CSRF_ENABLED=False,
    )
    return _init_userprofiles(app)


@pytest.fixture
def app_with_userprofiles_csrf(app):
    """Configure userprofiles module with CSRF enabled."""
    app.config.update(
        USERPROFILES_EXTEND_SECURITY_FORMS=True,
        WTF_CSRF_ENABLED=True,
    )
    return _init_userprofiles(app)


@pytest.fixture
def models_fixture(base_app):
    """Flask app with example data used to test models."""
    with base_app.app_context():
        datastore = base_app.extensions["security"].datastore
        datastore.create_user(
            email="existing@inveniosoftware.org", password="tester", active=True
        )
        datastore.create_user(
            email="test2@inveniosoftware.org", password="tester", active=True
        )
        datastore.create_user(
            email="test3@inveniosoftware.org", password="tester", active=True
        )
        datastore.commit()


@pytest.fixture
def params():
    """Fixture for remote app params."""

    def params(x):
        return dict(
            request_token_params={"scope": ""},
            base_url="https://foo.bar/",
            request_token_url=None,
            access_token_url="https://foo.bar/oauth/access_token",
            authorize_url="https://foo.bar/oauth/authorize",
            consumer_key=x,
            consumer_secret="testsecret",
            app_key="TEST_APP_CREDENTIALS",
        )

    return params


@pytest.fixture
def remote():
    """Fixture for remote app."""
    return type(
        "test_remote",
        (),
        dict(
            name="example_remote",
            request_token_params={"scope": ""},
            base_url="https://foo.bar/",
            request_token_url=None,
            access_token_url="https://foo.bar/oauth/access_token",
            authorize_url="https://foo.bar/oauth/authorize",
            consumer_key="testkey",
            consumer_secret="testsecret",
        ),
    )()


@pytest.fixture
def views_fixture(base_app, params, models_fixture):
    """Flask application with example data used to test views."""
    base_app.config["OAUTHCLIENT_REMOTE_APPS"].update(
        dict(
            test=dict(
                authorized_handler=lambda *args, **kwargs: "TEST",
                params=params("testid"),
                title="MyLinkedTestAccount",
            ),
            test_invalid=dict(
                authorized_handler=lambda *args, **kwargs: "TEST",
                params=params("test_invalidid"),
                title="Test Invalid",
            ),
            full=dict(
                params=params("fullid"),
                title="Full",
            ),
        )
    )

    return _init_app(base_app)


@pytest.fixture
def views_fixture_rest(base_app, params, models_fixture):
    """Flask application with example data used to test views."""
    base_app.config["OAUTHCLIENT_REST_REMOTE_APPS"].update(
        dict(
            test=dict(
                authorized_handler=lambda *args, **kwargs: "TEST",
                authorized_redirect_url="/",
                disconnect_redirect_url="/",
                signup_redirect_url="/",
                error_redirect_url="/",
                params=params("testid"),
                title="MyLinkedTestAccount",
            ),
            test_invalid=dict(
                authorized_handler=lambda *args, **kwargs: "TEST",
                authorized_redirect_url="/",
                disconnect_redirect_url="/",
                signup_redirect_url="/",
                error_redirect_url="/",
                params=params("test_invalidid"),
                title="Test Invalid",
            ),
            full=dict(
                params=params("fullid"),
                authorized_redirect_url="/",
                disconnect_redirect_url="/",
                signup_redirect_url="/",
                error_redirect_url="/",
                title="Full",
            ),
        )
    )


@pytest.fixture
def example_github(request):
    """ORCID example data."""
    return {
        "name": "Josiah Carberry",
        "expires_in": 3599,
        "access_token": "test_access_token",
        "refresh_token": "test_refresh_token",
        "scope": "/authenticate",
        "token_type": "bearer",
    }


@pytest.fixture
def example_globus(request):
    """Globus example data."""
    return (
        {
            "identity_provider_display_name": "Globus ID",
            "sub": "1142af3a-fea4-4df9-afe2-865ccd68bfdb",
            "preferred_username": "carberry@inveniosoftware.org",
            "identity_provider": "41143743-f3c8-4d60-bbdb-eeecaba85bd9",
            "organization": "Globus",
            "email": "carberry@inveniosoftware.org",
            "name": "Josiah Carberry",
        },
        {
            "expires_in": 3599,
            "resource_server": "auth.globus.org",
            "state": "test_state",
            "access_token": "test_access_token",
            "id_token": "header.test-oidc-token.pub-key",
            "other_tokens": [],
            "scope": "profile openid email",
            "token_type": "Bearer",
        },
        {
            "identities": [
                {
                    "username": "carberry@inveniosoftware.org",
                    "status": "used",
                    "name": "Josiah Carberry",
                    "email": "carberry@inveniosoftware.org",
                    "identity_provider": "927d7238-f917-4eb2-9ace-c523fa9ba34e",
                    "organization": "Globus",
                    "id": "3b843349-4d4d-4ef3-916d-2a465f9740a9",
                }
            ]
        },
    )


@pytest.fixture
def example_orcid(request):
    """ORCID example data."""
    return {
        "name": "Josiah Carberry",
        "expires_in": 3599,
        "orcid": "0000-0002-1825-0097",
        "access_token": "test_access_token",
        "refresh_token": "test_refresh_token",
        "scope": "/authenticate",
        "token_type": "bearer",
    }, dict(
        external_id="0000-0002-1825-0097",
        external_method="orcid",
        user=dict(profile=dict(full_name="Josiah Carberry")),
    )


@pytest.fixture()
def example_cern(request):
    """CERN example data."""
    file_path = os.path.join(
        os.path.dirname(__file__), "data/oauth_response_content.json"
    )
    with open(file_path) as response_file:
        json_data = response_file.read()

    return (
        OAuthResponse(resp=None, content=json_data, content_type="application/json"),
        dict(
            access_token="test_access_token",
            token_type="bearer",
            expires_in=1199,
            refresh_token="test_refresh_token",
        ),
        dict(
            user=dict(
                email="test.account@cern.ch",
                profile=dict(username="taccount", full_name="Test Account"),
            ),
            external_id="123456",
            external_method="cern",
            active=True,
        ),
    )


@pytest.fixture(scope="session")
def orcid_bio():
    """ORCID response fixture."""
    file_path = os.path.join(os.path.dirname(__file__), "data/orcid_bio.json")
    with open(file_path) as response_file:
        data = json.load(response_file)
    return data


@pytest.fixture()
def user(app_with_userprofiles):
    """Create users."""
    with db.session.begin_nested():
        datastore = app_with_userprofiles.extensions["security"].datastore
        user1 = datastore.create_user(
            email="info@inveniosoftware.org",
            password="tester",
            active=True,
            username="mynick",
        )
    db.session.commit()
    return user1


@pytest.fixture()
def user_rest(app_rest_with_userprofiles):
    """Create users."""
    with db.session.begin_nested():
        datastore = app_rest_with_userprofiles.extensions["security"].datastore
        user1 = datastore.create_user(
            email="info@inveniosoftware.org",
            password="tester",
            active=True,
            username="mynick",
        )
    db.session.commit()
    return user1


@pytest.fixture()
def form_test_data():
    """Test data to fill a registration form."""
    return dict(
        email="test@tester.com",
        profile=dict(
            full_name="Test Tester",
            username="test123",
        ),
    )


@pytest.fixture()
def example_keycloak_token():
    """Keycloak example data."""
    file_path = os.path.join(
        os.path.dirname(__file__), "data/keycloak_token_response.json"
    )

    with open(file_path) as token_file:
        token = json.load(token_file)

    return token


@pytest.fixture()
def example_keycloak_userinfo():
    """Keycloak example user info response."""
    file_path = os.path.join(
        os.path.dirname(__file__), "data/keycloak_userinfo_response.json"
    )

    with open(file_path) as response_file:
        response = json.load(response_file)

    return OAuthResponse(
        resp=None,
        content=json.dumps(response),
        content_type="application/json",
    )


@pytest.fixture()
def example_keycloak_realm_info():
    """Keycloak example realm info."""
    file_path = os.path.join(os.path.dirname(__file__), "data/keycloak_realm_info.json")

    with open(file_path) as info_file:
        realm_info = json.load(info_file)

    return realm_info


@pytest.fixture()
def example_keycloak_public_key():
    """Keycloak example public key."""
    return (
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAyZ5aoSIieF5vWOg4O8xTR"
        "lwWgs032Fuv7kvSXVLbE1XW+wlagD2asZ27U18A290RE3aK8zW4Bhtj8Zad/Xy1gA"
        "XYuuHNriNELBTm+WuJyA67bYvBB0QAKSIlDmbjc6btcgFJAyq0vNd0riwvsQJw5so"
        "RNT1eyrz12Z+yzCnSB5kO16ur2tzCxts+hkiUmznKcaPXDcbaJIUfyGMmSZLDk924"
        "GDFwOhLG01wcegmxUf48WqCQSfjzwudhubhJTSnNyhx8ndKUXKa3eGsV6Lub/u2di"
        "FZ+3rJGEbEKFUbFNPTJfslXh+mnH89/ZM8mZDb4V8YNX1lafSeJdvC7nnvvyQIDAQ"
        "AB"
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add Timestamp to oauthclient tables."""

from datetime import datetime

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "bff1f190b9bd"
down_revision = "97bbc733896c"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    current_date = datetime.utcnow()

    # Add 'created' and 'updated' columns to RemoteAccount
    _add_created_updated_columns("oauthclient_remoteaccount", current_date)

    # Add 'created' and 'updated' columns to RemoteToken
    _add_created_updated_columns("oauthclient_remotetoken", current_date)

    # Add 'created' and 'updated' columns to UserIdentity
    _add_created_updated_columns("oauthclient_useridentity", current_date)


def downgrade():
    """Downgrade database."""
    # Remove 'created' and 'updated' columns
    op.drop_column("oauthclient_remoteaccount", "created")
    op.drop_column("oauthclient_remoteaccount", "updated")

    op.drop_column("oauthclient_remotetoken", "created")
    op.drop_column("oauthclient_remotetoken", "updated")

    op.drop_column("oauthclient_useridentity", "created")
    op.drop_column("oauthclient_useridentity", "updated")


def _add_created_updated_columns(table, date):
    params = {"date": date}

    op.add_column(table, sa.Column("created", sa.DateTime()))
    op.add_column(table, sa.Column("updated", sa.DateTime()))

    op.execute(
        text("UPDATE " + table + " SET created= :date").bindparams(date=date), params
    )
    op.execute(
        text("UPDATE " + table + " SET updated= :date").bindparams(date=date), params
    )

    op.alter_column(table, "created", existing_type=sa.DateTime, nullable=False)
    op.alter_column(table, "updated", existing_type=sa.DateTime, nullable=False)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from functools import partial, wraps

import six
from flask import abort, current_app, session
from flask_login import current_user
from invenio_db import db
from werkzeug.utils import import_string

from ..errors import OAuthClientError, OAuthRejectedRequestError, OAuthResponseError
from ..models import RemoteAccount, RemoteToken
from ..proxies import current_oauthclient


#
# Token handling
#
def get_session_next_url(remote_app):
    """Return redirect url stored in session.

    :param remote_app: The remote application.
    :returns: The redirect URL.
    """
    return session.get("%s_%s" % (token_session_key(remote_app), "next_url"))


def set_session_next_url(remote_app, url):
    """Store redirect url in session for security reasons.

    :param remote_app: The remote application.
    :param url: the redirect URL.
    """
    session["%s_%s" % (token_session_key(remote_app), "next_url")] = url


def token_session_key(remote_app):
    """Generate a session key used to store the token for a remote app.

    :param remote_app: The remote application.
    :returns: The session key.
    """
    return "%s_%s" % (current_app.config["OAUTHCLIENT_SESSION_KEY_PREFIX"], remote_app)


def response_token_setter(remote, resp):
    """Extract token from response and set it for the user.

    :param remote: The remote application.
    :param resp: The response.
    :raises invenio_oauthclient.errors.OAuthClientError: If authorization with
        remote service failed.
    :raises invenio_oauthclient.errors.OAuthResponseError: In case of bad
        authorized request.
    :returns: The token.
    """
    if resp is None:
        raise OAuthRejectedRequestError("User rejected request.", remote, resp)
    else:
        if "access_token" in resp:
            return oauth2_token_setter(remote, resp)
        elif "oauth_token" in resp and "oauth_token_secret" in resp:
            return oauth1_token_setter(remote, resp)
        elif "error" in resp:
            # Only OAuth2 specifies how to send error messages
            raise OAuthClientError(
                "Authorization with remote service failed.",
                remote,
                resp,
            )
    raise OAuthResponseError("Bad OAuth authorized request", remote, resp)


def oauth1_token_setter(remote, resp, token_type="", extra_data=None):
    """Set an OAuth1 token.

    :param remote: The remote application.
    :param resp: The response.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance.
    """
    return token_setter(
        remote,
        resp["oauth_token"],
        secret=resp["oauth_token_secret"],
        extra_data=extra_data,
        token_type=token_type,
    )


def oauth2_token_setter(remote, resp, token_type="", extra_data=None):
    """Set an OAuth2 token.

    The refresh_token can be used to obtain a new access_token after
    the old one is expired. It is saved in the database for long term use.
    A refresh_token will be present only if `access_type=offline` is included
    in the authorization code request.

    :param remote: The remote application.
    :param resp: The response.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance.
    """
    return token_setter(
        remote,
        resp["access_token"],
        secret="",
        token_type=token_type,
        extra_data=extra_data,
    )


def token_setter(remote, token, secret="", token_type="", extra_data=None, user=None):
    """Set token for user.

    :param remote: The remote application.
    :param token: The token to set.
    :param token_type: The token type. (Default: ``''``)
    :param extra_data: Extra information. (Default: ``None``)
    :param user: The user owner of the remote token. If it's not defined,
        the current user is used automatically. (Default: ``None``)
    :returns: A :class:`invenio_oauthclient.models.RemoteToken` instance or
        ``None``.
    """
    session[token_session_key(remote.name)] = (token, secret)
    user = user or current_user

    # Save token if user is not anonymous (user exists but can be not active at
    # this moment)
    if not user.is_anonymous:
        uid = user.id
        cid = remote.consumer_key

        # Check for already existing token
        t = RemoteToken.get(uid, cid, token_type=token_type)

        if t:
            t.update_token(token, secret)
        else:
            t = RemoteToken.create(
                uid, cid, token, secret, token_type=token_type, extra_data=extra_data
            )
        return t
    return None


def token_getter(remote, token=""):
    """Retrieve OAuth access token.

    Used by flask-oauthlib to get the access token when making requests.

    :param remote: The remote application.
    :param token: Type of token to get. Data passed from ``oauth.request()`` to
        identify which token to retrieve. (Default: ``''``)
    :returns: The token.
    """
    session_key = token_session_key(remote.name)

    if session_key not in session and current_user.is_authenticated:
        # Fetch key from token store if user is authenticated, and the key
        # isn't already cached in the session.
        remote_token = RemoteToken.get(
            current_user.get_id(),
            remote.consumer_key,
            token_type=token,
        )

        if remote_token is None:
            return None

        # Store token and secret in session
        session[session_key] = remote_token.token()

    return session.get(session_key, None)


def token_delete(remote, token=""):
    """Remove OAuth access tokens from session.

    :param remote: The remote application.
    :param token: Type of token to get. Data passed from ``oauth.request()``
        to identify which token to retrieve. (Default: ``''``)
    :returns: The token.
    """
    session_key = token_session_key(remote.name)
    return session.pop(session_key, None)


def oauth_logout_handler(sender_app, user=None):
    """Remove all access tokens from session on logout."""
    oauth = current_oauthclient.oauth
    for remote in oauth.remote_apps.values():
        token_delete(remote)
    db.session.commit()


#
# Helpers
#
def make_handler(f, remote, with_response=True):
    """Make a handler for authorized and disconnect callbacks.

    :param f: Callable or an import path to a callable
    """
    if isinstance(f, six.string_types):
        f = import_string(f)

    @wraps(f)
    def inner(*args, **kwargs):
        if with_response:
            return f(args[0], remote, *args[1:], **kwargs)
        else:
            return f(remote, *args, **kwargs)

    return inner


def make_token_getter(remote):
    """Make a token getter for a remote application."""
    return partial(token_getter, remote)


def authorized_handler(f, authorized_response):
    """Handles an OAuth callback.

    As authorized_handler is deprecated in favor of authorized_response,
    it's has to be wrapped with handler which will be executed
    at the proper time.
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        data = authorized_response()
        return f(*((data,) + args), **kwargs)

    return decorated


def require_more_than_one_external_account(f):
    """Require that the user has more than one external account for login.

    If the user only has one linked external account and no means for logging
    in via local credentials, the decorated function won't be executed.
    This decorator is useful for disconnect handlers, to prevent users from
    disconnecting their last potential means of authentication.
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        if current_user.is_anonymous:
            return f(*args, **kwargs)

        local_login_enabled = current_app.config.get(
            "ACCOUNTS_LOCAL_LOGIN_ENABLED", True
        )
        password_set = current_user.password is not None
        local_login_possible = local_login_enabled and password_set

        remote_apps = current_app.config["OAUTHCLIENT_REMOTE_APPS"]
        accounts = RemoteAccount.query.filter_by(user_id=current_user.get_id()).all()

        # find out all of the linked external accounts for the user
        # that are currently configured and not hidden
        consumer_keys = set()
        remote_apps = current_app.config["OAUTHCLIENT_REMOTE_APPS"]
        for name, remote_app in remote_apps.items():
            if not remote_app.get("hide", False):
                consumer_keys.add(name)  # backcompat with v1.5.4
                remote_app_config = current_app.config[remote_app["params"]["app_key"]]
                consumer_keys.add(remote_app_config["consumer_key"])

        linked_accounts = [acc for acc in accounts if acc.client_id in consumer_keys]

        # execute the function only if local login is possible, or
        # there's more than one linked external account
        if local_login_possible or len(linked_accounts) > 1:
            return f(*args, **kwargs)

        else:
            abort(400)

    return decorated



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 TU Wien.
#
# Invenio-Keycloak is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for the Keycloak OAuth remote_app."""

import httpretty
import jwt
import pytest
from flask import session, url_for
from flask_login import current_user, login_user
from flask_security.utils import hash_password
from helpers import get_state, mock_keycloak
from invenio_accounts.models import User
from invenio_db import db
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient.contrib.keycloak.helpers import (
    _format_public_key,
    get_user_info,
)
from invenio_oauthclient.errors import OAuthError
from invenio_oauthclient.models import UserIdentity

# - - - - - - - - - - - - - -#
# Tests for Keycloak contrib #
# - - - - - - - - - - - - - -#


def test_login(app):
    """Test Keycloak login."""
    keycloak_config = app.config["OAUTHCLIENT_REMOTE_APPS"]["keycloak"]
    auth_url = keycloak_config["params"]["authorize_url"]

    client = app.test_client()

    resp = client.get(
        url_for("invenio_oauthclient.login", remote_app="keycloak", next="/someurl/")
    )

    assert resp.status_code == 302

    comps = urlparse(resp.location)
    params = parse_qs(comps.query)
    url = "{}://{}{}".format(comps.scheme, comps.netloc, comps.path)

    assert url == auth_url
    assert params["response_type"] == ["code"]
    assert params["scope"] == ["openid"]
    assert params["redirect_uri"]
    assert params["client_id"]
    assert params["state"]


@httpretty.activate
def test_authorized_signup_valid_user(
    app_with_userprofiles,
    example_keycloak_token,
    example_keycloak_userinfo,
    example_keycloak_realm_info,
):
    """Test authorized callback with sign-up."""
    app = app_with_userprofiles
    example_keycloak = example_keycloak_userinfo.data

    with app.test_client() as c:
        # ensure that remote_apps have been initialized (before first request)
        resp = c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))
        assert resp.status_code == 302

        # mock a running keycloak instance
        mock_keycloak(
            app.config,
            example_keycloak_token,
            example_keycloak,
            example_keycloak_realm_info,
        )

        # user authorized the request and is redirected back
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                code="test",
                state=get_state("keycloak"),
            )
        )

        # note: because we provided an e-mail address in 'info_handler',
        #       the user does not need to sign up
        assert resp.status_code == 302
        assert resp.location == "/account/settings/linkedaccounts/"

        # check that the user exists
        user = User.query.filter_by(email=example_keycloak["email"]).one()
        assert user is not None
        assert user.email == example_keycloak["email"]
        assert user.user_profile["full_name"] == "Max Moser"
        assert user.active
        assert user.confirmed_at

        # check that the user has a linked Keycloak account
        uid = UserIdentity.query.filter_by(
            method="keycloak", id_user=user.id, id=example_keycloak["sub"]
        ).one()
        assert uid.user is user

        # try to disconnect the Keycloak account again
        # which shouldn't work, because it's the user's only means of login
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="keycloak"))

        assert resp.status_code == 400

        # check that the user still exists
        user = User.query.filter_by(email=example_keycloak["email"]).one()
        assert user is not None

        # check that the Keycloak account hasn't been unlinked
        count = UserIdentity.query.filter_by(
            method="keycloak", id_user=user.id, id=example_keycloak["sub"]
        ).count()
        assert count == 1

        # set a password for the user
        user.password = hash_password("1234")
        db.session.commit()

        # try to disconnect the Keycloak account again
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="keycloak"))

        assert resp.status_code == 302

        # check that the user still exists
        user = User.query.filter_by(email=example_keycloak["email"]).one()
        assert user is not None

        # check that the Keycloak account hasn't been unlinked
        count = UserIdentity.query.filter_by(
            method="keycloak", id_user=user.id, id=example_keycloak["sub"]
        ).count()
        assert count == 0


@httpretty.activate
def test_authorized_signup_valid_user_not_confirmed(
    app_with_userprofiles,
    example_keycloak_token,
    example_keycloak_userinfo,
    example_keycloak_realm_info,
):
    """Test authorized callback with sign-up and user confirmed."""
    app = app_with_userprofiles
    example_keycloak = example_keycloak_userinfo.data

    # change default settings to avoid auto-confirm user
    keycloak_app = app.config["OAUTHCLIENT_REMOTE_APPS"]["keycloak"]
    keycloak_app["signup_options"]["auto_confirm"] = False
    keycloak_app["signup_options"]["send_register_msg"] = True

    with app.test_client() as c:
        # ensure that remote_apps have been initialized (before first request)
        resp = c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))
        assert resp.status_code == 302

        # mock a running keycloak instance
        mock_keycloak(
            app.config,
            example_keycloak_token,
            example_keycloak,
            example_keycloak_realm_info,
        )

        # user authorized the request and is redirected back
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                code="test",
                state=get_state("keycloak"),
            )
        )

        # note: because we provided an e-mail address in 'info_handler',
        #       the user does not need to sign up
        assert resp.status_code == 302
        assert resp.location == "/account/settings/linkedaccounts/"

        # check that the user has been created and confirmed
        user = User.query.filter_by(email=example_keycloak["email"]).one()
        assert user is not None
        assert user.email == example_keycloak["email"]
        assert user.user_profile["full_name"] == "Max Moser"
        assert user.active
        assert not user.confirmed_at


def test_authorized_reject(app, example_keycloak_token):
    """Test a rejected request."""
    with app.test_client() as c:
        c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))

        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                error="access_denied",
                error_description="User denied access",
                state=get_state("keycloak"),
            )
        )
        assert resp.status_code in (301, 302)
        assert resp.location == "/"

        # check message flash
        assert session["_flashes"][0][0] == "info"


@httpretty.activate
def test_authorized_already_authenticated(
    app,
    models_fixture,
    example_keycloak_token,
    example_keycloak_userinfo,
    example_keycloak_realm_info,
):
    """Test authorized callback with sign-in."""
    datastore = app.extensions["invenio-accounts"].datastore
    login_manager = app.login_manager

    example_keycloak = example_keycloak_userinfo.data
    existing_mail = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_mail)

    @login_manager.user_loader
    def load_user(user_id):
        return user

    @app.route("/logmein")
    def login():
        login_user(user)
        return "Logged in"

    with app.test_client() as c:
        c.get("/logmein", follow_redirects=True)

        # ensure that remote apps have been loaded (before first request)
        c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))

        # mock a running keycloak instance
        mock_keycloak(
            app.config,
            example_keycloak_token,
            example_keycloak,
            example_keycloak_realm_info,
        )

        # user goes to 'linked accounts' and clicks 'connect' with Keycloak
        resp = c.get(
            url_for(
                "invenio_oauthclient.login", remote_app="keycloak", next="/someurl/"
            )
        )

        assert resp.status_code == 302

        # the user logged in to Keycloak and authorized the request
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                code="test",
                state=get_state("keycloak"),
            )
        )

        # check if the Keycloak account has been linked to the user
        u = User.query.filter_by(email=existing_mail).one()
        UserIdentity.query.filter_by(
            method="keycloak", id_user=u.id, id=example_keycloak["sub"]
        ).one()

        # let the user hit the 'disconnect' button
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="keycloak"))
        assert resp.status_code == 302

        # check that the user still exists,
        # but the Keycloak account has been unlinked
        u = User.query.filter_by(email=existing_mail).one()
        count = UserIdentity.query.filter_by(
            method="keycloak", id_user=u.id, id=example_keycloak["sub"]
        ).count()
        assert count == 0


def test_not_authenticated(app):
    """Test disconnect when the user is not authenticated."""
    with app.test_client() as c:
        assert not current_user.is_authenticated
        resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="keycloak"))
        assert resp.status_code == 302


# - - - - - - - - - - - - - - - - - - #
# Tests for Keycloak helper functions #
# - - - - - - - - - - - - - - - - - - #


def test_format_public_key(example_keycloak_public_key):
    """Test if _format_public_key does roughly the right thing."""
    fmt_key = _format_public_key(example_keycloak_public_key)

    assert fmt_key != example_keycloak_public_key
    assert example_keycloak_public_key in fmt_key
    assert fmt_key.startswith("-----BEGIN PUBLIC KEY-----\n")
    assert fmt_key.endswith("\n-----END PUBLIC KEY-----")

    strip_key = fmt_key.lstrip("-----BEGIN PUBLIC KEY-----")
    strip_key = strip_key.rstrip("-----END PUBLIC KEY-----")
    strip_key = strip_key.strip()

    assert example_keycloak_public_key == strip_key


def test_format_public_key_idempotence(example_keycloak_public_key):
    """Test whether the _format_public_key function is idempotent."""
    fmt_once = _format_public_key(example_keycloak_public_key)
    fmt_twice = _format_public_key(fmt_once)
    fmt_thrice = _format_public_key(fmt_twice)

    assert fmt_once == fmt_twice == fmt_thrice


@httpretty.activate
def test_get_realm_key(
    app, example_keycloak_token, example_keycloak_userinfo, example_keycloak_realm_info
):
    """Test the mechanism for fetching the realm's public key."""
    mock_keycloak(
        app.config,
        example_keycloak_token,
        example_keycloak_userinfo.data,
        example_keycloak_realm_info,
    )


@httpretty.activate
def test_get_userinfo_from_token(
    app, example_keycloak_token, example_keycloak_userinfo, example_keycloak_realm_info
):
    """Test the "id_token" extraction mechanism from Keycloak's response."""
    mock_keycloak(
        app.config, example_keycloak_token, dict(), example_keycloak_realm_info
    )

    token = example_keycloak_token["id_token"]
    options = {"verify_signature": False}
    expected_result = jwt.decode(token, options=options)

    with app.test_client() as c:
        # ensure that remote apps have been loaded (before first request)
        c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))
        remote = app.extensions["oauthlib.client"].remote_apps["keycloak"]

        # the OAuthClient has to get its token from this call
        c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                code="test",
                state=get_state("keycloak"),
            )
        )

        # use token info
        user_info, _ = get_user_info(remote, example_keycloak_token)

        assert user_info is not None
        assert user_info == expected_result


@httpretty.activate
def test_get_userinfo_from_endpoint(
    app, example_keycloak_token, example_keycloak_userinfo, example_keycloak_realm_info
):
    """Test the "/userinfo" mechanism when the "id_token" mechanism fails."""
    mock_keycloak(
        app.config,
        example_keycloak_token,
        example_keycloak_userinfo.data,
        example_keycloak_realm_info,
    )

    app.config["OAUTHCLIENT_KEYCLOAK_USER_INFO_FROM_ENDPOINT"] = True

    with app.test_client() as c:
        # ensure that remote apps have been loaded (before first request)
        c.get(url_for("invenio_oauthclient.login", remote_app="keycloak"))

        remote = app.extensions["oauthlib.client"].remote_apps["keycloak"]

        # the OAuthClient has to get its token from this call
        c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="keycloak",
                code="test",
                state=get_state("keycloak"),
            )
        )

        # force the endpoint mechanism by not providing a token
        _, user_info = get_user_info(remote, None)

        assert user_info is not None
        assert user_info == example_keycloak_userinfo.data


def test_raise_on_invalid_app_name():
    """Test that the app name format is validated."""

    class FakeRemote:
        """Fake remote class."""

        def __init__(self, name):
            """Constructor."""
            self.name = name

    for invalid in ["i n v a l i d", "ke'c.ak"]:
        fr = FakeRemote(invalid)
        assert fr.name == invalid
        with pytest.raises(OAuthError):
            get_user_info(fr, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test utils."""

import sys

import pytest
from flask_security.confirmable import _security
from invenio_db import db
from six.moves.urllib.parse import quote_plus

from invenio_oauthclient.errors import AlreadyLinkedError
from invenio_oauthclient.models import RemoteAccount, RemoteToken
from invenio_oauthclient.proxies import current_oauthclient
from invenio_oauthclient.utils import (
    _get_external_id,
    create_csrf_disabled_registrationform,
    create_registrationform,
    fill_form,
    filter_user_info,
    get_safe_redirect_target,
    oauth_authenticate,
    oauth_get_user,
    oauth_link_external_id,
    oauth_unlink_external_id,
    obj_or_import_string,
    patch_dictionary,
    rebuild_access_tokens,
)


def test_utilities(app, models_fixture):
    """Test utilities."""
    datastore = app.extensions["invenio-accounts"].datastore
    assert obj_or_import_string("invenio_oauthclient.errors")

    # User
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    # Authenticate
    assert not _get_external_id({})
    assert not oauth_authenticate("dev", user, require_existing_link=True)

    _security.confirmable = True
    _security.login_without_confirmation = False
    user.confirmed_at = None
    assert not oauth_authenticate("dev", user)

    # Tokens
    t = RemoteToken.create(user.id, "dev", "mytoken", "mysecret")
    assert RemoteToken.get(
        user.id, "dev", access_token="mytoken"
    ) == RemoteToken.get_by_token("dev", "mytoken")

    assert oauth_get_user("dev", access_token=t.access_token) == user
    assert (
        oauth_get_user("dev", account_info={"user": {"email": existing_email}}) == user
    )

    # Link user to external id
    external_id = {"id": "123", "method": "test_method"}
    oauth_link_external_id(user, external_id)

    with pytest.raises(AlreadyLinkedError):
        oauth_link_external_id(user, external_id)

    assert (
        oauth_get_user(
            "dev",
            account_info={
                "external_id": external_id["id"],
                "external_method": external_id["method"],
            },
        )
        == user
    )

    # Cleanup
    oauth_unlink_external_id(external_id)
    acc = RemoteAccount.get(user.id, "dev")
    acc.delete()


def test_rebuilding_access_tokens(app, models_fixture):
    """Test rebuilding access tokens with random new SECRET_KEY."""
    old_secret_key = app.secret_key

    datastore = app.extensions["invenio-accounts"].datastore
    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    # Creating a new remote token and commiting to the db
    test_token = "mytoken"
    token_type = "testing"
    with db.session.begin_nested():
        rt = RemoteToken.create(
            user.id, "testkey", test_token, app.secret_key, token_type
        )
        db.session.add(rt)
    db.session.commit()

    # Changing application SECRET_KEY
    app.secret_key = "NEW_SECRET_KEY"
    db.session.expunge_all()

    # Asserting the decoding error occurs with the stale SECRET_KEY
    if sys.version_info[0] < 3:  # python 2
        remote_token = RemoteToken.query.first()
        assert remote_token.access_token != test_token
    else:  # python 3
        with pytest.raises(ValueError):
            RemoteToken.query.first()

    db.session.expunge_all()
    rebuild_access_tokens(old_secret_key)
    remote_token = RemoteToken.query.filter_by(token_type=token_type).first()

    # Asserting the access_token is not changed after rebuilding
    assert remote_token.access_token == test_token


def test_app_registrationform_missing_csrf(app, form_test_data):
    """App with CSRF disabled, registration form should not have it."""
    filled_form = _fill_form(app, create_registrationform, form_test_data)

    assert "profile" not in filled_form
    _assert_no_csrf_token(filled_form)


def test_app_registrationform_has_csrf(app_with_csrf, form_test_data):
    """App with CSRF enabled, test if registration form has it."""
    filled_form = _fill_form(app_with_csrf, create_registrationform, form_test_data)
    assert "profile" not in filled_form
    _assert_csrf_token(filled_form)


def test_registrationform_disable_csrf(app_with_csrf, form_test_data):
    """App with CSRF enabled, test if registration form removes it."""
    remote_apps = current_oauthclient.oauth.remote_apps
    first_remote_app = list(remote_apps.values())[0]
    filled_form = _fill_form(
        app_with_csrf,
        create_csrf_disabled_registrationform,
        form_test_data,
        remote=first_remote_app,
    )

    assert "profile" not in filled_form
    _assert_no_csrf_token(filled_form)


def test_registrationform_userprofile_missing_csrf(
    app_with_userprofiles, form_test_data
):
    """App with CSRF disabled and UserProfile, reg. form should not have it."""
    filled_form = _fill_form(
        app_with_userprofiles, create_registrationform, form_test_data
    )

    assert "profile" in filled_form
    assert "csrf_token" not in filled_form.profile
    _assert_no_csrf_token(filled_form)


def test_app_registrationform_userprofile_has_csrf(
    app_with_userprofiles_csrf, form_test_data
):
    """App with CSRF enabled and UserProfile, test if reg. form removes it."""
    filled_form = _fill_form(
        app_with_userprofiles_csrf, create_registrationform, form_test_data
    )

    assert "profile" in filled_form
    assert "csrf_token" not in filled_form.profile
    _assert_csrf_token(filled_form)


def test_registrationform_userprofile_disable_csrf(
    app_with_userprofiles_csrf, form_test_data
):
    """App with CSRF enabled and UserProfile, test if reg. form removes it."""
    remote_apps = current_oauthclient.oauth.remote_apps
    first_remote_app = list(remote_apps.values())[0]
    filled_form = _fill_form(
        app_with_userprofiles_csrf,
        create_csrf_disabled_registrationform,
        form_test_data,
        remote=first_remote_app,
    )

    assert "profile" in filled_form
    assert "csrf_token" not in filled_form.profile
    _assert_no_csrf_token(filled_form)


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "https://invenio.org/search?page=1&q=&keywords=taxonomy&keywords=animali",
            "/search?page=1&q=&keywords=taxonomy&keywords=animali",
        ),
        ("/search?page=1&size=20", "/search?page=1&size=20"),
        ("https://localhost/search?page=1", "https://localhost/search?page=1"),
        # previously encoded parameter
        (
            "/oauth/authorize?redirect_uri=http%3A%2F%2F127.0.0.1%3A5100%2Fauthorize",
            "/oauth/authorize?redirect_uri=http://127.0.0.1:5100/authorize",
        ),
    ],
)
def test_get_safe_redirect_target(app, test_input, expected):
    with app.test_request_context("/?next={0}".format(quote_plus(test_input))):
        assert get_safe_redirect_target() == expected


def _assert_csrf_token(form):
    """Assert that the field `csrf_token` exists in the form."""
    assert "csrf_token" in form
    assert form.csrf_token


def _assert_no_csrf_token(form):
    """Assert that the field `csrf_token` does not exist in the form."""

    # Flask-WTF==0.13.1 adds always `csrf_token` field, but with None value
    # Flask-WTF>0.14.2 do not `csrf_token` field
    assert "csrf_token" not in form or form.csrf_token.data is None


def _fill_form(app, form, data, *form_args, **form_kwargs):
    """Fill the input form with the provided data."""
    with app.test_request_context():
        filled_form = fill_form(form(*form_args, **form_kwargs), data)

        filled_form.validate()

        return filled_form


def test_patch_dictionary():
    """Test the dictionary patch function."""
    orig_dict = {
        "email": "user@inveniosoftware.org",
        "username": "user",
    }

    # patch some existing properties, add new ones, and leave some as is
    patch_dict = {
        "email": "admin@inveniosoftware.org",
        "profile": {
            "full_name": "Test User",
        },
        "extra": [1, 2, 3],
    }

    expected = {
        "email": "admin@inveniosoftware.org",
        "username": "user",
        "profile": {
            "full_name": "Test User",
        },
        "extra": [1, 2, 3],
    }

    patch_dictionary(orig_dict, patch_dict)
    assert orig_dict == expected


def test_precedence_mask(app):
    """Test if the precedence mask configuration is read properly."""
    precedence_mask = {
        "email": True,
        "password": False,
        "profile": {
            "username": True,
            "full_name": False,
            "extra1": True,
            "extra2": True,
            "extra3": False,
        },
    }

    user_info = {
        "email": "user@inveniosoftware.org",
        "password": "somepassword",
        "profile": {
            "username": "test-user",
            "full_name": "Test User",
            "extra2": 2,
            "extra3": 3,
            "extra4": 4,
        },
    }

    expected_filtered_user_info = {
        "email": "user@inveniosoftware.org",
        "profile": {
            "username": "test-user",
            "extra2": 2,
            "extra4": 4,
        },
    }

    input_values = {
        "email": "admin@inveniosoftware.org",
        "password": "anotherpassword",
        "profile": {
            "username": "admin",
            "full_name": "Ruler of the World",
            "extra2": 0,
            "extra3": 0,
        },
    }

    expected_values = {
        "email": "user@inveniosoftware.org",
        "password": "anotherpassword",
        "profile": {
            "username": "test-user",
            "full_name": "Ruler of the World",
            "extra2": 2,
            "extra3": 0,
            "extra4": 4,
        },
    }

    filter_user_info(user_info, precedence_mask)

    assert user_info == expected_filtered_user_info

    patch_dictionary(input_values, user_info)

    assert input_values == expected_values



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

r"""Minimal Flask application example for development with orcid handler.

SPHINX-START

1. Register an orcid application with `Authorization callback URL` as
   `http://localhost:5000/oauth/authorized/orcid/`

2. Install oauthclient:

   .. code-block:: console

      cdvirtualenv src/invenio-oauthclient
      pip install -e .[orcid]

3. Grab the *Client ID* and *Client Secret* after registering the application
   and add them to your instance configuration as `consumer_key` and
   `consumer_secret`.

   .. code-block:: console

       $ export ORCID_APP_CREDENTIALS_KEY=my_orcid_client_id
       $ export ORCID_APP_CREDENTIALS_SECRET=my_orcid_client_secret

4. Create database and tables:

   .. code-block:: console

       $ pip install -e .[all]
       $ cd examples
       $ export FLASK_APP=orcid_app.py
       $ ./app-setup.sh

You can find the database in `examples/orcid_app.db`.

5. Run the development server:

   .. code-block:: console

       $ flask -a orcid_app.py run -p 5000 -h '0.0.0.0'

6. Open in a browser the page `http://0.0.0.0:5000/orcid`.

   You will be redirected to orcid to authorize the application.

   Click on `Authorize application` and you will be redirected back to
   `http://0.0.0.0:5000/oauth/authorized/orcid/`, where you will be able to
   finalize the local user registration, inserting email address.

   Insert e.g. `fuu@bar.it` as email address and send the form.

   Now, you will be again in homepage but this time it say: `hello fuu@bar.it`.

   You have completed the user registration.

7. To be able to uninstall the example app:

   .. code-block:: console

       $ ./app-teardown.sh

SPHINX-END

"""

import os

from flask import Flask, redirect, url_for
from flask_login import current_user
from flask_menu import Menu as FlaskMenu
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint as blueprint_user
from invenio_db import InvenioDB
from invenio_i18n import Babel
from invenio_mail import InvenioMail as Mail
from invenio_userprofiles import InvenioUserProfiles
from invenio_userprofiles.views import (
    blueprint_api_init as blueprint_userprofile_api_init,
)
from invenio_userprofiles.views import (
    blueprint_ui_init as blueprint_userprofile_ui_init,
)

from invenio_oauthclient import InvenioOAuthClient
from invenio_oauthclient.contrib import orcid
from invenio_oauthclient.views.client import blueprint as blueprint_client
from invenio_oauthclient.views.settings import blueprint as blueprint_settings

from invenio_oauthclient._compat import monkey_patch_werkzeug  # noqa isort:skip

monkey_patch_werkzeug()  # noqa isort:skip

from flask_oauthlib.client import OAuth as FlaskOAuth  # noqa isort:skip

# [ Configure application credentials ]
ORCID_APP_CREDENTIALS = dict(
    consumer_key=os.environ.get("ORCID_APP_CREDENTIALS_KEY"),
    consumer_secret=os.environ.get("ORCID_APP_CREDENTIALS_SECRET"),
)

# Create Flask application
app = Flask(__name__)

app.config.update(
    SQLALCHEMY_ECHO=False,
    SQLALCHEMY_DATABASE_URI=os.environ.get(
        "SQLALCHEMY_DATABASE_URI", "sqlite:///orcid_app.db"
    ),
    OAUTHCLIENT_REMOTE_APPS=dict(
        orcid=orcid.REMOTE_SANDBOX_APP,
    ),
    ORCID_APP_CREDENTIALS=ORCID_APP_CREDENTIALS,
    DEBUG=True,
    SECRET_KEY="TEST",
    SECURITY_PASSWORD_SALT="security-password-salt",
    SECURITY_LOGIN_WITHOUT_CONFIRMATION=False,
    USERPROFILES_EXTEND_SECURITY_FORMS=True,
    SQLALCHEMY_TRACK_MODIFICATIONS=False,
    APP_THEME=["semantic-ui"],
    THEME_ICONS={"semantic-ui": dict(link="linkify icon")},
)

Babel(app)
FlaskMenu(app)
Mail(app)
InvenioDB(app)
InvenioAccounts(app)
InvenioUserProfiles(app)
FlaskOAuth(app)
InvenioOAuthClient(app)

app.register_blueprint(blueprint_user)
app.register_blueprint(blueprint_client)
app.register_blueprint(blueprint_settings)
app.register_blueprint(blueprint_userprofile_api_init)
app.register_blueprint(blueprint_userprofile_ui_init)


@app.route("/")
def index():
    """Homepage."""
    return "Home page (without any restrictions)"


@app.route("/orcid")
def orcid():
    """Try to print user email or redirect to login with orcid."""
    if not current_user.is_authenticated:
        return redirect(url_for("invenio_oauthclient.login", remote_app="orcid"))
    return "hello {}".format(current_user.email)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Views for OAuth."""

from flask_admin.contrib.sqla import ModelView
from invenio_accounts.admin import UserIdentityView, user_identity_adminview

from .models import RemoteAccount, RemoteToken


def _(x):
    """Identity."""
    return x


class RemoteAccountView(ModelView):
    """Flask-Admin view to manage remote accounts from invenio-oauthclient."""

    can_view_details = True

    column_list = (
        "id",
        "user_id",
        "client_id",
        "extra_data",
        "remote_tokens",
    )

    remote_account_columns = (
        "id",
        "user_id",
        "client_id",
        "extra_data",
    )

    column_searchable_list = column_sortable_list = remote_account_columns

    column_filters = (
        "id",
        "user_id",
        "client_id",
    )

    column_default_sort = ("id", True)

    column_display_all_relations = True
    inline_models = (RemoteToken,)

    column_labels = {
        "id": _("ID"),
        "user_id": _("User ID"),
        "client_id": _("Client ID"),
    }


class RemoteTokenView(ModelView):
    """Flask-Admin view to manage remote tokens from invenio-oauthclient."""

    can_view_details = True

    column_list = (
        "id_remote_account",
        "token_type",
    )

    column_searchable_list = column_sortable_list = column_list

    column_filters = (
        "id_remote_account",
        "token_type",
    )

    form_columns = (
        "remote_account",
        "token_type",
    )

    column_labels = {
        "id_remote_account": _("ID Remote Account"),
    }


remote_account_adminview = {
    "model": RemoteAccount,
    "modelview": RemoteAccountView,
    "category": _("User Management"),
    "name": _("Linked accounts"),
}


remote_token_adminview = {
    "model": RemoteToken,
    "modelview": RemoteTokenView,
    "category": _("User Management"),
    "name": _("Linked account tokens"),
}

__all__ = (
    "remote_account_adminview",
    "remote_token_adminview",
    "user_identity_adminview",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 University of Chicago.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pre-configured remote application for enabling sign in/up with Globus.

1. Edit your configuration and add:

   .. code-block:: python

        from invenio_oauthclient.contrib import globus

        OAUTHCLIENT_REMOTE_APPS = dict(
            globus=globus.REMOTE_APP,
        )

        GLOBUS_APP_CREDENTIALS = dict(
            consumer_key='changeme',
            consumer_secret='changeme',
        )

2. Register a Globus application at `https://developers.globus.org/` with the
   `Redirect URL` as `http://localhost:5000/oauth/authorized/globus/`. For
   full documentation on all app fields, see:
   `https://docs.globus.org/api/auth/developer-guide/#register-app`

4. Grab the *Client ID* and *Client Secret* after registering the application
   and add them to your instance configuration (``invenio.cfg``):

   .. code-block:: python

        GLOBUS_APP_CREDENTIALS = dict(
            consumer_key='<CLIENT ID>',
            consumer_secret='<CLIENT SECRET>',
        )

5. Now go to your site: http://localhost:5000/oauth/authorized/globus/

6. You should see Globus listed under Linked accounts:
   http://localhost:5000/account/settings/linkedaccounts/

In case you would prefer a different title and description for this app
you can re-define the default Globus OAuth instance:

.. code-block:: python

        from invenio_oauthclient.contrib import globus

        _my_app = GlobusOAuthSettingsHelper("Globus provider",
                                            "another description")
        _REMOTE_APP = _my_app.remote_app

        OAUTHCLIENT_REMOTE_APPS = dict(
            globus=_REMOTE_APP,
        )

        GLOBUS_APP_CREDENTIALS = dict(
            consumer_key='changeme',
            consumer_secret='changeme',
        )
"""

from flask import current_app, redirect, url_for
from flask_login import current_user
from invenio_db import db

from invenio_oauthclient import current_oauthclient
from invenio_oauthclient.contrib.settings import OAuthSettingsHelper
from invenio_oauthclient.errors import OAuthResponseError
from invenio_oauthclient.handlers.rest import response_handler
from invenio_oauthclient.handlers.utils import require_more_than_one_external_account
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import oauth_link_external_id, oauth_unlink_external_id


class GlobusOAuthSettingsHelper(OAuthSettingsHelper):
    """Default configuration for Globus OAuth provider."""

    def __init__(
        self,
        title=None,
        description=None,
        base_url=None,
        app_key=None,
        precedence_mask=None,
        signup_options=None,
    ):
        """Constructor."""
        super().__init__(
            title or "Globus",
            description or "Research data management simplified.",
            base_url or "https://auth.globus.org/v2/",
            app_key or "GLOBUS_APP_CREDENTIALS",
            request_token_params={"scope": "openid email profile"},
            precedence_mask=precedence_mask,
            signup_options=signup_options,
        )

        self._handlers = dict(
            authorized_handler="invenio_oauthclient.handlers:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.globus:disconnect_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.globus:account_info",
                info_serializer="invenio_oauthclient.contrib.globus:account_info_serializer",
                setup="invenio_oauthclient.contrib.globus:account_setup",
                view="invenio_oauthclient.handlers:signup_handler",
            ),
        )

        self._rest_handlers = dict(
            authorized_handler="invenio_oauthclient.handlers.rest:authorized_signup_handler",
            disconnect_handler="invenio_oauthclient.contrib.globus:disconnect_rest_handler",
            signup_handler=dict(
                info="invenio_oauthclient.contrib.globus:account_info",
                info_serializer="invenio_oauthclient.contrib.globus:account_info_serializer",
                setup="invenio_oauthclient.contrib.globus:account_setup",
                view="invenio_oauthclient.handlers.rest:signup_handler",
            ),
            response_handler="invenio_oauthclient.handlers.rest:default_remote_response_handler",
            authorized_redirect_url="/",
            disconnect_redirect_url="/",
            signup_redirect_url="/",
            error_redirect_url="/",
        )

    def get_handlers(self):
        """Return Globus auth handlers."""
        return self._handlers

    def get_rest_handlers(self):
        """Return Globus auth REST handlers."""
        return self._rest_handlers

    @property
    def user_info_url(self):
        """Return the URL to fetch user info."""
        return f"{self.base_url}oauth2/userinfo"

    @property
    def user_identity_url(self):
        """Return the URL to fetch user identity."""
        return f"{self.base_url}api/identities"


_globus_app = GlobusOAuthSettingsHelper()

BASE_APP = _globus_app.base_app
GLOBUS_USER_INFO_URL = _globus_app.user_info_url
GLOBUS_USER_ID_URL = _globus_app.user_identity_url
GLOBUS_EXTERNAL_METHOD = "globus"
"""Kept only for backward compat, they should not be used."""

REMOTE_APP = _globus_app.remote_app
"""Globus remote application configuration."""
REMOTE_REST_APP = _globus_app.remote_rest_app
"""Globus remote rest application configuration."""


def get_dict_from_response(response):
    """Check for errors in the response and return the resulting JSON."""
    if getattr(response, "_resp") and response._resp.code > 400:
        raise OAuthResponseError(
            "Application mis-configuration in Globus", None, response
        )

    return response.data


def get_user_info(remote):
    """Get user information from Globus.

    See the docs here for v2/oauth/userinfo:
    https://docs.globus.org/api/auth/reference/
    """
    response = remote.get(_globus_app.user_info_url)
    user_info = get_dict_from_response(response)
    response.data["username"] = response.data["preferred_username"]
    if "@" in response.data["username"]:
        user_info["username"], _ = response.data["username"].split("@")
    return user_info


def get_user_id(remote, email):
    """Get the Globus identity for the given email.

    A Globus ID is a UUID that can uniquely identify a Globus user. See the
    docs here for v2/api/identities
    https://docs.globus.org/api/auth/reference/
    """
    try:
        url = f"{remote.base_url}oauth2/userinfo?usernames={email}"
        user_id = get_dict_from_response(remote.get(url))
        return user_id["identities"][0]["id"]
    except KeyError:
        # If we got here the response was successful but the data was invalid.
        # It's likely the URL is wrong but possible the API has changed.
        raise OAuthResponseError(
            "Failed to fetch user id, likely server " "mis-configuration", None, remote
        )


def account_info_serializer(remote, resp, user_info, user_id, **kwargs):
    """Serialize the account info response object.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :param user_info: The response of the `user info` endpoint.
    :param user_id: The user id.
    :returns: A dictionary with serialized user information.
    """
    return {
        "user": {
            "email": user_info["email"],
            "profile": {
                "username": user_info["username"],
                "full_name": user_info["name"],
            },
        },
        "external_id": user_id,
        "external_method": remote.name,
    }


def account_info(remote, resp):
    """Retrieve remote account information used to find local user.

    It returns a dictionary with the following structure:

    .. code-block:: python

        {
            'user': {
                'email': '...',
                'profile': {
                    'username': '...',
                    'full_name': '...',
                }
            },
            'external_id': 'globus-unique-identifier',
            'external_method': 'globus',
        }

    Information inside the user dictionary are available for other modules.
    For example, they are used from the module invenio-userprofiles to fill
    the user profile.

    :param remote: The remote application.
    :param resp: The response of the `authorized` endpoint.
    :returns: A dictionary with the user information.
    """
    user_info = get_user_info(remote)
    user_id = get_user_id(remote, user_info["preferred_username"])

    handlers = current_oauthclient.signup_handlers[remote.name]
    # `remote` param automatically injected via `make_handler` helper
    return handlers["info_serializer"](resp, user_info, user_id=user_id)


def account_setup(remote, token, resp):
    """Perform additional setup after user have been logged in.

    :param remote: The remote application.
    :param token: The token value.
    :param resp: The response.
    """
    info = get_user_info(remote)
    user_id = get_user_id(remote, info["preferred_username"])
    with db.session.begin_nested():
        token.remote_account.extra_data = {"login": info["username"], "id": user_id}

        # Create user <-> external id link.
        oauth_link_external_id(
            token.remote_account.user,
            dict(id=user_id, method=remote.name),
        )


@require_more_than_one_external_account
def _disconnect(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    if not current_user.is_authenticated:
        return current_app.login_manager.unauthorized()

    remote_account = RemoteAccount.get(
        user_id=current_user.get_id(), client_id=remote.consumer_key
    )
    external_ids = [
        i.id for i in current_user.external_identifiers if i.method == remote.name
    ]

    if external_ids:
        oauth_unlink_external_id(dict(id=external_ids[0], method=remote.name))

    if remote_account:
        with db.session.begin_nested():
            remote_account.delete()


def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    _disconnect(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


def disconnect_rest_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    :param remote: The remote application.
    :returns: The HTML response.
    """
    _disconnect(remote, *args, **kwargs)
    redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][remote.name][
        "disconnect_redirect_url"
    ]
    return response_handler(remote, redirect_url)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility methods to help find, authenticate or register a remote user."""

from datetime import datetime

from flask import after_this_request, current_app, request
from flask_security import login_user, logout_user
from flask_security.confirmable import confirm_user, requires_confirmation
from invenio_accounts.models import User
from invenio_accounts.utils import register_user
from invenio_db.utils import rebuild_encrypted_properties
from itsdangerous import TimedJSONWebSignatureSerializer
from uritools import uricompose, urisplit
from werkzeug.local import LocalProxy
from werkzeug.utils import import_string

from .models import RemoteAccount, RemoteToken, UserIdentity

_security = LocalProxy(lambda: current_app.extensions["security"])

_datastore = LocalProxy(lambda: _security.datastore)


serializer = LocalProxy(
    lambda: TimedJSONWebSignatureSerializer(
        current_app.config["SECRET_KEY"],
        expires_in=current_app.config["OAUTHCLIENT_STATE_EXPIRES"],
    )
)


def _commit(response=None):
    _datastore.commit()
    return response


def _get_external_id(account_info):
    """Get external id from account info."""
    if all(k in account_info for k in ("external_id", "external_method")):
        return dict(
            id=account_info["external_id"], method=account_info["external_method"]
        )
    return None


def oauth_get_user(client_id, account_info=None, access_token=None):
    """Retrieve user object for the given request.

    Uses either the access token or extracted account information to retrieve
    the user object.

    :param client_id: The client id.
    :param account_info: The dictionary with the account info.
        (Default: ``None``)
    :param access_token: The access token. (Default: ``None``)
    :returns: A :class:`invenio_accounts.models.User` instance or ``None``.
    """
    if access_token:
        token = RemoteToken.get_by_token(client_id, access_token)
        if token:
            return token.remote_account.user

    if account_info:
        external_id = _get_external_id(account_info)
        if external_id:
            user = UserIdentity.get_user(external_id["method"], external_id["id"])
            if user:
                return user
        email = account_info.get("user", {}).get("email")
        if email:
            return User.query.filter_by(email=email).one_or_none()
    return None


def oauth_authenticate(client_id, user, require_existing_link=False):
    """Authenticate an oauth authorized callback.

    :param client_id: The client id.
    :param user: A user instance.
    :param require_existing_link: If ``True``, check if remote account exists.
        (Default: ``False``)
    :returns: ``True`` if the user is successfully authenticated.
    """
    # Authenticate via the access token (access token used to get user_id)
    if not requires_confirmation(user):
        after_this_request(_commit)
        if login_user(user):
            if require_existing_link:
                account = RemoteAccount.get(user.id, client_id)
                if account is None:
                    logout_user()
                    return False
            return True
    return False


def filter_user_info(user_info, precedence_mask):
    """Filter the user info dictionary according to the precedence mask.

    :param user_info: The user info dictionary.
    :param precedence_mask: The precedence mask to use.
    """
    # for each of the user info values, check if they are supposed
    # to take precedence over user input (as per precedence mask)
    for key, user_info_value in list(user_info.items()):
        precedence_value = precedence_mask.get(key, True)
        info_val_dict = isinstance(user_info_value, dict)
        prec_val_dict = isinstance(precedence_value, dict)

        if info_val_dict and prec_val_dict:
            # if both values in the mask and user_info are dicts,
            # investigate deeper
            filter_user_info(user_info_value, precedence_mask[key])

        elif prec_val_dict:
            # the precedence mask says it's a dict, but it is actually
            # a different value... remove this inconsistent user_info value
            user_info.pop(key, None)

        elif not precedence_value:
            user_info.pop(key, None)


def patch_dictionary(orig_dict, patch_dict):
    """Replace the fields mentioned in the patch, while leaving others as is.

    Note: The first argument's content will be changed during the process.

    :param orig_dict: A dictionary.
    :param patch_dict: The dictionary whose values should take precedence.
    """
    for key in patch_dict.keys():
        val = patch_dict[key]
        orig_val = orig_dict.get(key)

        if isinstance(val, dict) and isinstance(orig_val, dict):
            patch_dictionary(orig_val, val)
        else:
            orig_dict[key] = val


def remove_csrf_tokens(user_data):
    """Remove CSRF tokens from the user data."""
    user_data.pop("csrf_token", None)
    for key, value in list(user_data.items()):
        if isinstance(value, dict):
            remove_csrf_tokens(value)


def remove_none_values(user_data):
    """Remove None values from the user data."""
    del_keys = []
    for key, value in list(user_data.items()):
        if isinstance(value, dict):
            remove_none_values(value)
            if value == {}:
                del_keys.append(key)
        if value is None:
            del_keys.append(key)
    for key in del_keys:
        del user_data[key]


def oauth_register(form, user_info=None, precedence_mask=None, signup_options={}):
    """Register user if possible.

    :param form: A form instance.
    :param user_info: The user info dictionary.
    :param precedence_mask: The precedence mask to use.
    :param signup_options: Extra signup options dict.
    :returns: A :class:`invenio_accounts.models.User` instance.
    """
    if form.validate():
        form_data = form.data

        # let relevant information from the OAuth service's user info
        # have precedence over the values specified by the user
        if user_info:
            default_mask = {"email": True}
            # act on form data so the `profile` is updated correctly
            filter_user_info(user_info, precedence_mask or default_mask)
            patch_dictionary(form_data, user_info)

        # re-populate the form after applying the precedence mask and
        # convert the form data to user model's data
        data = fill_form(form, form_data).to_dict()
        # see https://github.com/inveniosoftware/invenio-oauthclient/issues/275
        remove_none_values(data)
        # remove the CSRF tokens to avoid unexpected keyword arguments
        remove_csrf_tokens(data)

        send_register_msg = signup_options.get("send_register_msg", True)
        user = register_user(send_register_msg=send_register_msg, **data)
        auto_confirm = signup_options.get("auto_confirm", False)
        if auto_confirm:
            confirm_user(user)
        _datastore.commit()
        return user


def oauth_link_external_id(user, external_id=None):
    """Link a user to an external id.

    :param user: A :class:`invenio_accounts.models.User` instance.
    :param external_id: The external id associated with the user.
        (Default: ``None``)
    :raises invenio_oauthclient.errors.AlreadyLinkedError: Raised if already
        exists a link.
    """
    # Backward compatibility. Use UserIdentity directly instead of this method.
    UserIdentity.create(user, external_id["method"], external_id["id"])


def oauth_unlink_external_id(external_id):
    """Unlink a user from an external id.

    :param external_id: The external id associated with the user.
    """
    UserIdentity.delete_by_external_id(external_id["method"], external_id["id"])


def get_safe_redirect_target(arg="next"):
    """Get URL to redirect to and ensure that it is local.

    :param arg: URL argument.
    :returns: The redirect target or ``None``.
    """
    allowed_hosts = current_app.config.get("APP_ALLOWED_HOSTS") or []
    for target in request.args.get(arg), request.referrer:
        if target:
            redirect_uri = urisplit(target)
            if redirect_uri.host in allowed_hosts:
                return target
            elif redirect_uri.path:
                return uricompose(
                    path=redirect_uri.getpath(),
                    query=redirect_uri.getquery(),
                    fragment=redirect_uri.getfragment(),
                )
    return None


def obj_or_import_string(value, default=None):
    """Import string or return object."""
    if isinstance(value, str):
        return import_string(value)
    elif value:
        return value
    return default


def load_or_import_from_config(key, app=None, default=None):
    """Load or import value from config."""
    app = app or current_app
    imp = app.config.get(key)
    return obj_or_import_string(imp, default=default)


def _create_registrationform(*args, **kwargs):
    """Default registration form after external auth success."""

    class RegistrationForm(_security.confirm_register_form):
        password = None
        recaptcha = None
        submit = None  # defined in the template

    return RegistrationForm(*args, **kwargs)


def create_registrationform(*args, **kwargs):
    """Make a registration form."""
    func = current_app.config["OAUTHCLIENT_SIGNUP_FORM"]
    return func(*args, **kwargs)


def create_csrf_disabled_registrationform(remote):
    """Create a registration form with CSRF disabled."""
    func = current_app.config["OAUTHCLIENT_SIGNUP_FORM"]
    return func(oauth_remote_app=remote, **_get_csrf_disabled_param())


def fill_form(form, data):
    """Prefill form with data.

    :param form: The form to fill.
    :param data: The data to insert in the form.
    :returns: A pre-filled form.
    """
    for key, value in data.items():
        if hasattr(form, key):
            field = getattr(form, key)
            if isinstance(value, dict):
                fill_form(field, value)
            elif field is not None:
                field.data = value
    return form


def rebuild_access_tokens(old_key):
    """Rebuild the access token field when the SECRET_KEY is changed.

    Fixes users' login

    :param old_key: the old SECRET_KEY.
    """
    current_app.logger.info("rebuilding RemoteToken.access_token...")
    rebuild_encrypted_properties(old_key, RemoteToken, ["access_token"])


def _get_csrf_disabled_param():
    """Return the right param to disable CSRF depending on WTF-Form version.

    From Flask-WTF 0.14.0, `csrf_enabled` param has been deprecated in favor of
    `meta={csrf: True/False}`.
    """
    import flask_wtf
    from pkg_resources import parse_version

    supports_meta = parse_version(flask_wtf.__version__) >= parse_version("0.14.0")
    return dict(meta={"csrf": False}) if supports_meta else dict(csrf_enabled=False)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module level errors."""

# Imported here for backward compatibility after moving UserIdentity to
# Invenio-Accounts.
from invenio_accounts.errors import AlreadyLinkedError


class OAuthError(Exception):
    """Base class for OAuth exceptions."""

    def __init__(self, message, remote):
        """Initialize exception.

        :param message: Error message.
        :param message: Remote application.
        """
        self.message = message
        self.remote = remote


class OAuthResponseError(OAuthError):
    """Define response exception during OAuth process."""

    def __init__(self, message, remote, response):
        """Initialize exception.

        :param message: Error message.
        :param remote: Remote application.
        :param response: OAuth response object.
        """
        super().__init__(message, remote)
        self.response = response


class OAuthRejectedRequestError(OAuthResponseError):
    """Define exception for rejected response during OAuth process."""


class OAuthClientError(OAuthResponseError):
    """Define OAuth client exception.

    Client errors happens when the client (i.e. Invenio) creates an invalid
    request.
    """

    def __init__(self, message, remote, response):
        """Initialize exception.

        :param message: Error message.
        :param remote: Remote application.
        :param response: OAuth response object. Used to extract ``error``,
                         ``error_uri`` and ``error_description``.
        """
        # Only OAuth2 specifies how to send error messages
        self.code = response["error"]
        self.uri = response.get("error_uri", None)
        self.description = response.get("error_description", None)
        super().__init__(self.description or message, remote, response)


class OAuthCERNRejectedAccountError(OAuthResponseError):
    """Define exception for not allowed cern accounts."""


class OAuthKeycloakUserInfoError(OAuthResponseError):
    """Define exception for problems while fetching user info from Keycloak."""


class OAuthClientUnAuthorized(Exception):
    """Define exception for unauthorized user."""


class OAuthClientAlreadyAuthorized(Exception):
    """Define exception for user already authorized."""


class OAuthClientTokenNotFound(Exception):
    """Define exception for oauth token not found."""


class OAuthClientUserNotRegistered(Exception):
    """Define exception for user not registered."""


class OAuthClientTokenNotSet(Exception):
    """Define exception for oauth token not set."""


class OAuthClientMustRedirectSignup(Exception):
    """Define exception for forcing redirection to signup view."""


class OAuthClientMustRedirectLogin(Exception):
    """Define exception for forcing redirection to login view."""


class OAuthRemoteNotFound(Exception):
    """Define exception for remote app not found."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for github oauth remote app."""

from collections import namedtuple

import mock
import pytest
from flask import session, url_for
from flask_login import current_user
from flask_security import login_user
from flask_security.utils import hash_password
from helpers import check_redirect_location, mock_response
from invenio_accounts.models import User
from invenio_db import db
from six.moves.urllib_parse import parse_qs, urlparse

from invenio_oauthclient._compat import _create_identifier
from invenio_oauthclient.contrib.github import authorized
from invenio_oauthclient.errors import OAuthResponseError
from invenio_oauthclient.models import RemoteAccount, RemoteToken, UserIdentity
from invenio_oauthclient.views.client import serializer


def _get_state():
    return serializer.dumps(
        {
            "app": "github",
            "sid": _create_identifier(),
            "next": None,
        }
    )


def test_login(app):
    """Test github login."""
    client = app.test_client()

    resp = client.get(
        url_for("invenio_oauthclient.login", remote_app="github", next="/someurl/")
    )
    assert resp.status_code == 302

    params = parse_qs(urlparse(resp.location).query)
    assert params["response_type"], ["code"]
    assert params["scope"] == ["user,user:email"]
    assert params["redirect_uri"]
    assert params["client_id"]
    assert params["state"]


class MockGh(object):
    """Mock gh."""

    def __init__(self, email):
        """Init."""
        self._email = email

    def emails(self):
        """Get emails."""
        Email = namedtuple("Email", "verified primary email")
        yield Email(verified=True, primary=True, email=self._email)

    def me(self):
        """Mock me."""
        Me = namedtuple("Me", "id name login")
        return Me(id="githubuser", name="John", login="mynick")


def test_authorized_signup_valid_user(app, example_github):
    """Test authorized callback with sign-up."""
    example_email = "info@inveniosoftware.org"

    with app.test_client() as c:
        # User login with email 'info'
        with mock.patch("github3.login") as MockLogin:
            MockLogin.return_value = MockGh(email="info@inveniosoftware.org")

            # Ensure remote apps have been loaded (due to before first
            # request)
            resp = c.get(url_for("invenio_oauthclient.login", remote_app="github"))

            assert resp.status_code == 302

            mock_response(app.extensions["oauthlib.client"], "github", example_github)

            # User authorized the requests and is redirect back
            resp = c.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="github",
                    code="test",
                    state=_get_state(),
                )
            )
            assert resp.status_code == 302
            assert resp.location == "/account/settings/linkedaccounts/"

            # Assert database state (Sign-up complete)
            user = User.query.filter_by(email=example_email).one()
            remote = RemoteAccount.query.filter_by(user_id=user.id).one()
            RemoteToken.query.filter_by(id_remote_account=remote.id).one()
            assert user.active
            assert user.confirmed_at

            # Disconnect link
            # should not work, because it's the user's only means of login
            resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="github"))
            assert resp.status_code == 400

            user = User.query.filter_by(email=example_email).one()
            assert (
                1
                == UserIdentity.query.filter_by(
                    method="github", id_user=user.id, id="githubuser"
                ).count()
            )

            # set a password for the user
            user.password = hash_password("1234")
            db.session.commit()

            # Disconnect again
            resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="github"))
            assert resp.status_code == 302

            user = User.query.filter_by(email=example_email).one()
            assert (
                0
                == UserIdentity.query.filter_by(
                    method="github", id_user=user.id, id="githubuser"
                ).count()
            )

            assert RemoteAccount.query.filter_by(user_id=user.id).count() == 0
            assert RemoteToken.query.count() == 0

        # User login with another email ('info2')
        with mock.patch("github3.login") as MockLogin:
            MockLogin.return_value = MockGh(email="info2@inveniosoftware.org")

            # User authorized the requests and is redirect back
            resp = c.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="github",
                    code="test",
                    state=_get_state(),
                )
            )
            assert resp.status_code == 302
            assert resp.location == "/account/settings/linkedaccounts/"

            # check that exist only one account
            user = User.query.filter_by(email=example_email).one()
            assert User.query.count() == 1


def test_authorized_signup_username_already_exists(app, example_github, user):
    """Test authorized callback with sign-up."""
    example_email = "another@email.it"

    with app.test_client() as c:
        # User login with email 'info'
        with mock.patch("github3.login") as MockLogin:
            MockLogin.return_value = MockGh(email=example_email)

            # Ensure remote apps have been loaded (due to before first
            # request)
            resp = c.get(url_for("invenio_oauthclient.login", remote_app="github"))

            assert resp.status_code == 302

            mock_response(app.extensions["oauthlib.client"], "github", example_github)

            # User authorized the requests and is redirect back
            resp = c.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="github",
                    code="test",
                    state=_get_state(),
                )
            )
            assert resp.status_code == 302
            assert resp.location == (
                url_for("invenio_oauthclient.signup", remote_app="github")
            )

            # User fills form to register
            resp = c.post(
                resp.headers["Location"],
                data={
                    "email": example_email,
                    "password": "123456",
                    "profile.username": "pippo2",
                    "profile.full_name": "pluto",
                    "profile.affiliations": "CERN",
                },
            )
            assert resp.status_code == 302

            # Assert database state (Sign-up complete)
            my_user = User.query.filter_by(email=example_email).one()
            remote = RemoteAccount.query.filter_by(user_id=my_user.id).one()
            RemoteToken.query.filter_by(id_remote_account=remote.id).one()
            assert my_user.active

            # Disconnect link
            # should not work, because it's the user's only means of login
            resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="github"))
            assert resp.status_code == 400

            my_user = User.query.filter_by(email=example_email).one()
            assert (
                1
                == UserIdentity.query.filter_by(
                    method="github", id_user=my_user.id, id="githubuser"
                ).count()
            )

            # set a password for the user
            my_user.password = hash_password("1234")
            db.session.commit()

            # Disconnect again
            resp = c.get(url_for("invenio_oauthclient.disconnect", remote_app="github"))
            assert resp.status_code == 302

            my_user = User.query.filter_by(email=example_email).one()
            assert (
                0
                == UserIdentity.query.filter_by(
                    method="github", id_user=my_user.id, id="githubuser"
                ).count()
            )
            assert RemoteAccount.query.filter_by(user_id=my_user.id).count() == 0
            assert RemoteToken.query.count() == 0
            assert User.query.count() == 2


def test_authorized_reject(app):
    """Test a rejected request."""
    with app.test_client() as c:
        c.get(url_for("invenio_oauthclient.login", remote_app="github"))
        resp = c.get(
            url_for(
                "invenio_oauthclient.authorized",
                remote_app="github",
                error="access_denied",
                error_description="User denied access",
                state=_get_state(),
            )
        )
        assert resp.status_code in (301, 302)
        assert resp.location == "/"
        # Check message flash
        assert session["_flashes"][0][0] == "info"


def test_authorized_already_authenticated(app, models_fixture, example_github):
    """Test authorized callback with sign-up."""
    datastore = app.extensions["invenio-accounts"].datastore
    login_manager = app.login_manager

    existing_email = "existing@inveniosoftware.org"
    user = datastore.find_user(email=existing_email)

    @login_manager.user_loader
    def load_user(user_id):
        return user

    @app.route("/foo_login")
    def login():
        login_user(user)
        return "Logged In"

    with mock.patch("github3.login") as MockLogin:
        MockLogin.return_value = MockGh(email="info@inveniosoftware.org")

        with app.test_client() as client:
            # make a fake login (using my login function)
            client.get("/foo_login", follow_redirects=True)

            # Ensure remote apps have been loaded (due to before first
            # request)
            client.get(url_for("invenio_oauthclient.login", remote_app="github"))

            # Mock access token request
            mock_response(app.extensions["oauthlib.client"], "github", example_github)

            # User then goes to 'Linked accounts' and clicks 'Connect'
            resp = client.get(
                url_for(
                    "invenio_oauthclient.login", remote_app="github", next="/someurl/"
                )
            )
            assert resp.status_code == 302

            # User authorized the requests and is redirected back
            resp = client.get(
                url_for(
                    "invenio_oauthclient.authorized",
                    remote_app="github",
                    code="test",
                    state=_get_state(),
                )
            )

            # Assert database state (Sign-up complete)
            u = User.query.filter_by(email=existing_email).one()
            remote = RemoteAccount.query.filter_by(user_id=u.id).one()
            RemoteToken.query.filter_by(id_remote_account=remote.id).one()

            # Disconnect link
            resp = client.get(
                url_for("invenio_oauthclient.disconnect", remote_app="github")
            )
            assert resp.status_code == 302

            # User exists
            u = User.query.filter_by(email=existing_email).one()
            assert (
                0
                == UserIdentity.query.filter_by(
                    method="github", id_user=u.id, id="githubuser"
                ).count()
            )
            assert RemoteAccount.query.filter_by(user_id=u.id).count() == 0
            assert RemoteToken.query.count() == 0


def test_not_authenticated(app):
    """Test disconnect when user is not authenticated."""
    with app.test_client() as client:
        assert not current_user.is_authenticated
        resp = client.get(
            url_for("invenio_oauthclient.disconnect", remote_app="github")
        )
        assert resp.status_code == 302


def test_authorized_handler(app, remote):
    """Test authorized callback handler."""
    # General error
    example_response = {"error": "error"}
    resp = authorized(example_response, remote)
    check_redirect_location(resp, "/")

    # Bad verification error
    example_response = {"error": "bad_verification_code"}
    resp = authorized(example_response, remote)
    check_redirect_location(resp, "/oauth/login/github/")

    # Incorrect client credentials
    example_response = {"error": "incorrect_client_credentials"}
    with pytest.raises(OAuthResponseError):
        authorized(example_response, remote)

        # Redirect uri mismatch
        example_response = {"error": "redirect_uri_mismatch"}
        with pytest.raises(OAuthResponseError):
            authorized(example_response, remote)
